#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module running the plot command
========================

Exposes the interface for plotting for command-line usage.

"""
import logging

import orion.core.io.experiment_builder as experiment_builder
from orion.client.experiment import ExperimentClient
from orion.core.cli import base as cli
from orion.plotting.base import SINGLE_EXPERIMENT_PLOTS

log = logging.getLogger(__name__)
DESCRIPTION = "Produce plots for Oríon experiments"


IMAGE_TYPES = ["png", "jpg", "jpeg", "webp", "svg", "pdf"]
HTML_TYPES = ["html"]
JSON_TYPES = ["json"]
VALID_TYPES = IMAGE_TYPES + HTML_TYPES + JSON_TYPES


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    plot_parser = parser.add_parser("plot", help=DESCRIPTION, description=DESCRIPTION)

    cli.get_basic_args_group(plot_parser)

    plot_parser.add_argument(
        "kind",
        type=str,
        choices=SINGLE_EXPERIMENT_PLOTS.keys(),
        help="kind of plot to generate. ",
    )

    plot_parser.add_argument(
        "-t",
        "--type",
        type=str,
        default="png",
        choices=VALID_TYPES,
        help="type of plot to return. (default: png)",
    )

    plot_parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="",
        help="path where plot is saved. "
        " Will override the `type` argument."
        " (default is {exp.name}-v{exp.version}_{kind}.{type})",
    )

    plot_parser.add_argument(
        "--scale",
        type=float,
        default=1,
        help="more pixels, but same proportions of the plot. "
        " Scale acts as multiplier on height and width of resulting image."
        " Overrides value of 'scale' in plotly.io.write_image.",
    )

    plot_parser.set_defaults(func=main)

    return plot_parser


def infer_type(output, out_type):
    """Infer type of plot file based on output filename or provided type.

    If output has a valid extension, this extension is used as the type. Otherwise,
    the provided (or default) output type is used.
    """
    if output:
        ext = output.split(".")[-1]
        if ext and ext not in VALID_TYPES:
            log.warning(
                "Overriding the `type` field with something from `output`, "
                f"but we got the invalid value : {out_type}. Will revert to png."
                f"Should be one of {VALID_TYPES}"
            )
        else:
            out_type = ext

    return out_type


def get_output(experiment, output, kind, out_type):
    """Create output file name based on experiment name, plot kind and file type.

    If the output filename is provided, it is appended with the file type if filename
    does not already has the corresponding extention. (ex output.name -> output.name.png)
    """

    if not output:
        return f"{experiment.name}-v{experiment.version}_{kind}.{out_type}"
    elif not output.endswith(f".{out_type}"):
        return f"{output}.{out_type}"

    return output


def main(args):
    """Starts an application that will generate a plot."""

    # Note : If you specify no argument at all (except 'kind'),
    #        the default behavior is to plot "{experiment.name}_{kind}.png".

    experiment = ExperimentClient(
        experiment_builder.get_from_args(args, mode="r"), None
    )
    output_plot = experiment.plot(kind=args["kind"])

    args["type"] = infer_type(args["output"], args["type"])
    args["output"] = get_output(experiment, args["output"], args["kind"], args["type"])

    if args["type"] in IMAGE_TYPES:
        output_plot.write_image(args["output"], scale=args["scale"])
    elif args["type"] in HTML_TYPES:
        output_plot.write_html(args["output"])
    elif args["type"] in JSON_TYPES:
        with open(args["output"], "w") as f_out:
            # Note that this is the content of the "body" in the WebApi.
            f_out.write(output_plot.to_json())
    else:
        raise Exception(
            "This is a bug. You should never land here if the logic is not faulty."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.experiment_branch_builder`."""

import copy
import os

import pytest
import yaml

import orion.core
import orion.core.utils.backward as backward
from orion.core import evc
from orion.core.evc.conflicts import (
    AlgorithmConflict,
    ChangedDimensionConflict,
    CodeConflict,
    CommandLineConflict,
    ExperimentNameConflict,
    MissingDimensionConflict,
    NewDimensionConflict,
    OrionVersionConflict,
    ScriptConfigConflict,
    detect_conflicts,
)
from orion.core.io.experiment_branch_builder import ExperimentBranchBuilder


def filter_true(c):
    """Filter solved conflicts"""
    return c.is_resolved is True


def filter_false(c):
    """Filter unsolved conflicts"""
    return not filter_true(c)


@pytest.fixture
def user_config():
    """Generate data dict for user's script's configuration file"""
    data = {
        "a": "orion~uniform(-10,10)",
        "some_other": "test",
        "b": "orion~normal(0,1)",
        "argument": "value",
    }
    return data


@pytest.fixture
def parent_config(user_config):
    """Create a configuration that will not hit the database."""
    user_script = "tests/functional/demo/black_box.py"
    config = dict(
        _id="test",
        name="test",
        algorithms="random",
        version=1,
        metadata={
            "VCS": {
                "type": "git",
                "is_dirty": False,
                "HEAD_sha": "test",
                "active_branch": None,
                "diff_sha": "diff",
            },
            "user_script": user_script,
            "user_args": [
                user_script,
                "--nameless=option",
                "-x~uniform(0,1)",
                "-y~normal(0,1)",
                "-z~uniform(0,10)",
                "--manual-resolution",
            ],
            "user": "some_user_name",
            "orion_version": "XYZ",
        },
        refers={},
    )

    config_file_path = "./parent_config.yaml"

    with open(config_file_path, "w") as f:
        yaml.dump(user_config, f)

    config["metadata"]["user_args"].append("--config=%s" % config_file_path)

    backward.populate_space(config)

    yield config
    os.remove(config_file_path)


@pytest.fixture
def child_config(parent_config, storage):
    """Create a child branching from the test experiment"""
    config = copy.deepcopy(parent_config)
    config["_id"] = "test2"
    config["version"] = 2
    config["refers"]["parent_id"] = parent_config["_id"]
    return config


@pytest.fixture
def missing_config(child_config):
    """Create a child config with a missing dimension"""
    del child_config["metadata"]["user_args"][2]  # del -x
    del child_config["metadata"]["user_args"][2]  # del -y
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def new_config(child_config):
    """Create a child config with a new dimension"""
    child_config["metadata"]["user_args"].append("-w_d~normal(0,1)")
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def changed_config(child_config):
    """Create a child config with a changed dimension"""
    second_element = child_config["metadata"]["user_args"][3]
    second_element = second_element.replace("normal", "uniform")
    child_config["metadata"]["user_args"][3] = second_element
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def changed_algo_config(child_config):
    """Create a child config with a new algo"""
    child_config["algorithms"] = "stupid-grid"
    return child_config


@pytest.fixture
def changed_orion_version_config(child_config):
    """Create a child config with new orion version"""
    child_config["metadata"]["orion_version"] = "UVW"
    return child_config


@pytest.fixture
def changed_code_config(child_config):
    """Create a child config with a changed dimension"""
    child_config["metadata"]["VCS"]["HEAD_sha"] = "new_test"
    return child_config


@pytest.fixture
def same_userconfig_config(user_config, child_config):
    """Create a child config with a changed dimension"""
    config_file_path = "./same_config.yaml"
    with open(config_file_path, "w") as f:
        yaml.dump(user_config, f)
    child_config["metadata"]["user_args"][-1] = "--config=%s" % config_file_path
    backward.populate_space(child_config)
    yield child_config
    os.remove(config_file_path)


@pytest.fixture
def changed_userconfig_config(user_config, child_config):
    """Create a child config with a changed dimension"""
    config_file_path = "./changed_config.yaml"
    user_config["b"] = "orion~uniform(-20, 0, precision=None)"
    user_config["some_other"] = "hello"
    with open(config_file_path, "w") as f:
        yaml.dump(user_config, f)
    child_config["metadata"]["user_args"][-1] = "--config=%s" % config_file_path
    backward.populate_space(child_config)
    yield child_config
    os.remove(config_file_path)


@pytest.fixture
def changed_cli_config(child_config):
    """Create a child config with a changed dimension"""
    child_config["metadata"]["user_args"] += ["-u=0", "--another=test", "positional"]
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def list_arg_with_equals_cli_config(child_config):
    """Create a child config with an argument of the
    form --args=1 --args=2 --args=3
    """
    child_config["metadata"]["user_args"] += ["--args=1", "--args=2", "--args=3"]
    backward.populate_space(child_config)
    return child_config


@pytest.fixture
def cl_config():
    """Create a child config with markers for commandline solving"""
    user_script = "tests/functional/demo/black_box.py"
    config = dict(
        name="test",
        branch="test2",
        algorithms="random",
        metadata={
            "hash_commit": "old",
            "user_script": user_script,
            "user_args": [
                user_script,
                "--nameless=option",
                "-x~>w_d",
                "-w_d~+normal(0,1)",
                "-y~+uniform(0,1)",
                "-z~-",
                "--omega~+normal(0,1)",
            ],
            "user": "some_user_name",
            "orion_version": "XYZ",
        },
    )
    backward.populate_space(config)
    return config


@pytest.fixture
def conflicts(
    new_dimension_conflict,
    changed_dimension_conflict,
    missing_dimension_conflict,
    algorithm_conflict,
    orion_version_conflict,
    code_conflict,
    experiment_name_conflict,
    config_conflict,
    cli_conflict,
):
    """Create a container for conflicts with one of each types for testing purposes"""
    conflicts = evc.conflicts.Conflicts()
    conflicts.register(new_dimension_conflict)
    conflicts.register(changed_dimension_conflict)
    conflicts.register(missing_dimension_conflict)
    conflicts.register(algorithm_conflict)
    conflicts.register(orion_version_conflict)
    conflicts.register(code_conflict)
    conflicts.register(experiment_name_conflict)
    conflicts.register(config_conflict)
    conflicts.register(cli_conflict)
    return conflicts


class TestConflictDetection(object):
    """Test detection of conflicts between two configurations"""

    def test_no_conflicts(self, parent_config, child_config):
        """Test the case where the child is the same as the parent"""
        conflicts = detect_conflicts(parent_config, child_config)

        assert len(conflicts.get()) == 1
        assert isinstance(conflicts.get()[0], ExperimentNameConflict)

    def test_missing_dim_conflict(self, parent_config, missing_config):
        """Test if missing dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, missing_config)

        assert len(conflicts.get()) == 3
        conflict = conflicts.get()[1]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == "/x"
        assert isinstance(conflict, MissingDimensionConflict)

    def test_new_dim_conflict(self, parent_config, new_config):
        """Test if new dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, new_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[1]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == "/w_d"
        assert isinstance(conflict, NewDimensionConflict)

    def test_changed_dim_conflict(self, parent_config, changed_config):
        """Test if changed dimension is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == "/y"
        assert isinstance(conflict, ChangedDimensionConflict)

    def test_changed_dim_userconfig_conflict(
        self, parent_config, changed_userconfig_config
    ):
        """Test if changed dimension from user's config is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)

        assert len(conflicts.get()) == 4
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.dimension.name == "/b"
        assert isinstance(conflict, ChangedDimensionConflict)

    def test_algo_conflict(self, parent_config, changed_algo_config):
        """Test if algorithm changes are currently detected"""
        conflicts = detect_conflicts(parent_config, changed_algo_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.old_config["algorithms"] == "random"
        assert conflict.new_config["algorithms"] == "stupid-grid"
        assert isinstance(conflict, AlgorithmConflict)

    def test_orion_version_conflict(self, parent_config, changed_orion_version_config):
        """Test if orion version changes are currently detected"""
        conflicts = detect_conflicts(parent_config, changed_orion_version_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[1]

        assert conflict.is_resolved is False
        assert conflict.old_config["metadata"]["orion_version"] == "XYZ"
        assert conflict.new_config["metadata"]["orion_version"] == "UVW"
        assert isinstance(conflict, OrionVersionConflict)

    def test_code_conflict(self, parent_config, changed_code_config):
        """Test if code commit hash change is currently detected"""
        conflicts = detect_conflicts(parent_config, changed_code_config)

        assert len(conflicts.get()) == 2
        conflict = conflicts.get()[0]

        assert conflict.is_resolved is False
        assert conflict.old_config["metadata"]["VCS"]["HEAD_sha"] == "test"
        assert conflict.new_config["metadata"]["VCS"]["HEAD_sha"] == "new_test"
        assert isinstance(conflict, CodeConflict)

    def test_ignore_code_conflict(self, parent_config, changed_code_config):
        """Test if ignored code commit hash change is detected as a conflict"""
        conflicts = detect_conflicts(
            parent_config, changed_code_config, {"ignore_code_changes": True}
        )

        assert len(conflicts.get()) == 1

    def test_config_new_name_no_conflict(self, parent_config, same_userconfig_config):
        """Test if same configuration file with a different name is not detected as a conflict"""
        conflicts = detect_conflicts(parent_config, same_userconfig_config)

        assert parent_config["metadata"]["user_args"][-1].startswith("--config=")
        assert same_userconfig_config["metadata"]["user_args"][-1].startswith(
            "--config="
        )
        assert (
            parent_config["metadata"]["user_args"][-1]
            != same_userconfig_config["metadata"]["user_args"][-1]
        )

        assert len(conflicts.get()) == 2
        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved

    def test_config_non_dim_conflict(self, parent_config, changed_userconfig_config):
        """Test if changed configuration file is detected as a conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)

        assert len(conflicts.get()) == 4
        assert not conflicts.get([ChangedDimensionConflict])[0].is_resolved
        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert not conflicts.get([ScriptConfigConflict])[0].is_resolved

    def test_cli_conflict(self, parent_config, changed_cli_config):
        """Test if changed command line call is detected as a conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)

        assert len(conflicts.get()) == 2

        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert not conflicts.get([CommandLineConflict])[0].is_resolved

    def test_cli_ignored_conflict(self, parent_config, changed_cli_config):
        """Test if ignored changed command line call is detected as a conflict"""
        changed_cli_config["metadata"]["user_args"].pop()
        conflicts = detect_conflicts(
            parent_config,
            changed_cli_config,
            {"non_monitored_arguments": ["u", "another"]},
        )

        assert len(conflicts.get()) == 1

        assert not conflicts.get([ExperimentNameConflict])[0].is_resolved


class TestResolutions(object):
    """Test resolution of conflicts"""

    def test_add_single_hit(self, parent_config, new_config):
        """Test if adding a dimension only touches the correct status"""
        del new_config["metadata"]["user_args"][2]
        backward.populate_space(new_config)
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension("w_d")

        assert len(conflicts.get()) == 3
        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[0].is_resolved

    def test_add_new(self, parent_config, new_config):
        """Test if adding a new dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension("w_d")

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)

    def test_add_changed(self, parent_config, changed_config):
        """Test if adding a changed dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, changed_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.add_dimension("y")

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ChangeDimensionResolution)

    def test_remove_missing(self, parent_config, missing_config):
        """Test if removing a missing dimension solves the conflict"""
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.remove_dimension("x")

        assert len(conflicts.get()) == 3
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)

    def test_rename_missing(self, parent_config, missing_config):
        """Test if renaming a dimension to another solves both conflicts"""
        missing_config["metadata"]["user_args"].append("-w_d~uniform(0,1)")
        backward.populate_space(missing_config)
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        branch_builder.rename_dimension("x", "w_d")

        assert len(conflicts.get()) == 4

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert conflicts.get([MissingDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[1].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(
            resolved_conflicts[1].resolution,
            resolved_conflicts[1].RenameDimensionResolution,
        )
        assert resolved_conflicts[1].resolution.conflict.dimension.name == "/x"
        assert (
            resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name
            == "/w_d"
        )

    def test_rename_missing_changed(self, parent_config, missing_config):
        """Test if renaming a dimension to another with different prior solves both conflicts but
        creates a new one which is not solved
        """
        missing_config["metadata"]["user_args"].append("-w_d~normal(0,1)")
        backward.populate_space(missing_config)
        conflicts = detect_conflicts(parent_config, missing_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4

        branch_builder.rename_dimension("x", "w_d")

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get([NewDimensionConflict])[0].is_resolved
        assert conflicts.get([MissingDimensionConflict])[0].is_resolved
        assert not conflicts.get([MissingDimensionConflict])[1].is_resolved
        assert not conflicts.get([ChangedDimensionConflict])[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(
            resolved_conflicts[1].resolution,
            resolved_conflicts[1].RenameDimensionResolution,
        )
        assert resolved_conflicts[1].resolution.conflict.dimension.name == "/x"
        assert (
            resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name
            == "/w_d"
        )

    def test_reset_dimension(self, parent_config, new_config):
        """Test if resetting a dimension unsolves the conflict"""
        conflicts = detect_conflicts(parent_config, new_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        branch_builder.add_dimension("w_d")
        assert len(conflicts.get_resolved()) == 2

        with pytest.raises(ValueError) as exc:
            branch_builder.reset("w_d~+")
        assert "'w_d~+' is not in list" in str(exc.value)
        assert len(conflicts.get_resolved()) == 2

        branch_builder.reset("w_d~+norm(0, 1)")

        assert len(conflicts.get()) == 2

        conflict = conflicts.get(dimension_name="w_d")[0]

        assert not conflict.is_resolved
        assert isinstance(conflict, NewDimensionConflict)
        assert len(conflicts.get_resolved()) == 1

    def test_name_experiment(
        self, bad_exp_parent_config, bad_exp_child_config, storage
    ):
        """Test if having the same experiment name does not create a conflict."""
        backward.populate_space(bad_exp_parent_config)
        backward.populate_space(bad_exp_child_config)
        storage.create_experiment(bad_exp_parent_config)
        storage.create_experiment(bad_exp_child_config)
        conflicts = detect_conflicts(bad_exp_parent_config, bad_exp_parent_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

        conflict = conflicts.get([ExperimentNameConflict])[0]

        assert conflict.new_config["name"] == "test"
        assert not conflict.is_resolved
        branch_builder.change_experiment_name("test2")
        assert len(conflicts.get_resolved()) == 1
        assert conflict.new_config["name"] == "test2"
        assert conflict.is_resolved

    def test_bad_name_experiment(self, parent_config, child_config, monkeypatch):
        """Test if changing the experiment names does not work for invalid name and revert
        to old one
        """

        def _is_unique(self, *args, **kwargs):
            return False

        def _versions(self, *args, **kwargs):
            return True

        monkeypatch.setattr(
            ExperimentNameConflict.ExperimentNameResolution,
            "_name_is_unique",
            _is_unique,
        )
        monkeypatch.setattr(
            ExperimentNameConflict.ExperimentNameResolution,
            "_check_for_greater_versions",
            _versions,
        )

        conflicts = detect_conflicts(parent_config, child_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

        conflict = conflicts.get([ExperimentNameConflict])[0]

        assert not conflict.is_resolved
        branch_builder.change_experiment_name("test2")
        assert len(conflicts.get_resolved()) == 0
        assert conflict.new_config["name"] == "test"
        assert not conflict.is_resolved

    def test_algo_change(self, parent_config, changed_algo_config):
        """Test if setting the algorithm conflict solves it"""
        conflicts = detect_conflicts(parent_config, changed_algo_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_algo()

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AlgorithmResolution)

    def test_orion_version_change(self, parent_config, changed_orion_version_config):
        """Test if setting the orion version conflict solves it"""
        conflicts = detect_conflicts(parent_config, changed_orion_version_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_orion_version()

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.OrionVersionResolution)

    def test_code_change(self, parent_config, changed_code_config):
        """Test if giving a proper change-type solves the code conflict"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_code_change_type(evc.adapters.CodeChange.types[0])

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CodeConflict)

    def test_bad_code_change(self, capsys, parent_config, changed_code_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_code_change_type("bad-type")
        out, err = capsys.readouterr()
        assert "Invalid code change type" in out.split("\n")[-3]

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if giving a proper change-type solves the user script config conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_script_config_change_type(
            evc.adapters.ScriptConfigChange.types[0]
        )

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]
        assert conflict.is_resolved
        assert isinstance(conflict, ScriptConfigConflict)

    def test_bad_config_change(self, capsys, parent_config, changed_userconfig_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_script_config_change_type("bad-type")
        out, err = capsys.readouterr()
        assert "Invalid script's config change type" in out.split("\n")[-3]

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 1

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if giving a proper change-type solves the command line conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        branch_builder.set_cli_change_type(evc.adapters.CommandLineChange.types[0])

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CommandLineConflict)

    def test_bad_cli_change(self, capsys, parent_config, changed_cli_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)
        capsys.readouterr()
        branch_builder.set_cli_change_type("bad-type")
        out, err = capsys.readouterr()
        assert "Invalid cli change type" in out.split("\n")[-3]

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

    def test_solve_all_automatically(self, conflicts):
        """Test if all conflicts all automatically resolve by the ExperimentBranchBuilder."""
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get_resolved()) == 9


class TestResolutionsWithMarkers(object):
    """Test resolution of conflicts with markers"""

    def test_add_new(self, parent_config, new_config):
        """Test if new dimension conflict is automatically resolved"""
        new_config["metadata"]["user_args"][-1] = "-w_d~+normal(0,1)"
        conflicts = detect_conflicts(parent_config, new_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)

    def test_add_new_default(self, parent_config, new_config):
        """Test if new dimension conflict is automatically resolved"""
        new_config["metadata"]["user_args"][-1] = "-w_d~+normal(0,1,default_value=0)"
        backward.populate_space(new_config)
        conflicts = detect_conflicts(parent_config, new_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AddDimensionResolution)
        assert conflict.resolution.default_value == 0.0

    def test_add_bad_default(self, parent_config, new_config):
        """Test if new dimension conflict raises an error if marked with invalid default value"""
        new_config["metadata"]["user_args"][-1] = "-w_d~+normal(0,1,default_value='a')"
        backward.populate_space(new_config)
        with pytest.raises(TypeError) as exc:
            detect_conflicts(parent_config, new_config)
        assert "Parameter '/w_d': Incorrect arguments." in str(exc.value)

    def test_add_changed(self, parent_config, changed_config):
        """Test if changed dimension conflict is automatically resolved"""
        changed_config["metadata"]["user_args"][3] = changed_config["metadata"][
            "user_args"
        ][3].replace("~", "~+")
        conflicts = detect_conflicts(parent_config, changed_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ChangeDimensionResolution)

    def test_remove_missing(self, parent_config, child_config):
        """Test if missing dimension conflict is automatically resolved"""
        child_config["metadata"]["user_args"][2] = "-x~-"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)

    def test_remove_missing_default(self, parent_config, child_config):
        """Test if missing dimension conflict is automatically resolved"""
        child_config["metadata"]["user_args"][2] = "-x~-0.5"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.RemoveDimensionResolution)
        assert conflict.resolution.default_value == 0.5

    def test_remove_missing_bad_default(self, parent_config, child_config):
        """Test if missing dimension conflict raises an error if marked with invalid default"""
        child_config["metadata"]["user_args"][2] = "-x~--100"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[1]

        assert not conflict.is_resolved
        assert isinstance(conflict, MissingDimensionConflict)

    def test_rename_missing(self, parent_config, child_config):
        """Test if renaming is automatically applied with both conflicts resolved"""
        child_config["metadata"]["user_args"].append("-w_a~uniform(0,1)")
        child_config["metadata"]["user_args"].append("-w_b~normal(0,1)")
        child_config["metadata"]["user_args"][2] = "-x~>w_a"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name="x")[0].is_resolved
        assert conflicts.get(dimension_name="w_a")[0].is_resolved
        assert not conflicts.get(dimension_name="w_b")[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(
            resolved_conflicts[1].resolution,
            resolved_conflicts[1].RenameDimensionResolution,
        )
        assert resolved_conflicts[1].resolution.conflict.dimension.name == "/x"
        assert (
            resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name
            == "/w_a"
        )

    def test_rename_invalid(self, parent_config, child_config):
        """Test if renaming to invalid dimension raises an error"""
        child_config["metadata"]["user_args"].append("-w_a~uniform(0,1)")
        child_config["metadata"]["user_args"].append("-w_b~uniform(0,1)")
        child_config["metadata"]["user_args"][2] = "-x~>w_c"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        with pytest.raises(ValueError) as exc:
            ExperimentBranchBuilder(conflicts, manual_resolution=True)
        assert "Dimension name 'w_c' not found in conflicts" in str(exc.value)

    def test_rename_missing_changed(self, parent_config, child_config):
        """Test if renaming is automatically applied with both conflicts resolved,
        but not the new one because of prior change
        """
        child_config["metadata"]["user_args"].append("-w_a~uniform(0,1)")
        child_config["metadata"]["user_args"].append("-w_b~normal(0,1)")
        child_config["metadata"]["user_args"][2] = "-x~>w_b"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name="x")[0].is_resolved
        assert conflicts.get([NewDimensionConflict], dimension_name="w_b")[
            0
        ].is_resolved
        assert not conflicts.get([ChangedDimensionConflict], dimension_name="w_b")[
            0
        ].is_resolved
        assert not conflicts.get(dimension_name="w_a")[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 3
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(
            resolved_conflicts[1].resolution,
            resolved_conflicts[1].RenameDimensionResolution,
        )
        assert resolved_conflicts[1].resolution.conflict.dimension.name == "/x"
        assert (
            resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name
            == "/w_b"
        )

    def test_rename_missing_changed_marked(self, parent_config, child_config):
        """Test if renaming is automatically applied with all conflicts resolved including
        the new one caused by prior change
        """
        child_config["metadata"]["user_args"].append("-w_a~uniform(0,1)")
        child_config["metadata"]["user_args"].append("-w_b~+normal(0,1)")
        child_config["metadata"]["user_args"][2] = "-x~>w_b"
        backward.populate_space(child_config)
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 5

        assert conflicts.get([ExperimentNameConflict])[0].is_resolved
        assert conflicts.get(dimension_name="x")[0].is_resolved
        assert conflicts.get([NewDimensionConflict], dimension_name="w_b")[
            0
        ].is_resolved
        assert conflicts.get([ChangedDimensionConflict], dimension_name="w_b")[
            0
        ].is_resolved
        assert not conflicts.get(dimension_name="w_a")[0].is_resolved

        resolved_conflicts = conflicts.get_resolved()
        assert len(resolved_conflicts) == 4
        assert resolved_conflicts[1].resolution is resolved_conflicts[2].resolution
        assert isinstance(
            resolved_conflicts[1].resolution,
            resolved_conflicts[1].RenameDimensionResolution,
        )
        assert resolved_conflicts[1].resolution.conflict.dimension.name == "/x"
        assert (
            resolved_conflicts[1].resolution.new_dimension_conflict.dimension.name
            == "/w_b"
        )

    def test_name_experiment_version_update(self, parent_config, child_config, storage):
        """Test if experiment name conflict is automatically resolved with version update"""
        old_name = "test"
        new_version = 2
        storage.create_experiment(parent_config)
        child_config["version"] = 1
        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[0]

        assert conflict.resolution.new_name == old_name
        assert conflict.resolution.new_version == 2
        assert conflict.new_config["name"] == old_name
        assert conflict.new_config["version"] == new_version
        assert conflict.is_resolved

    def test_name_experiment_name_change(self, parent_config, child_config, storage):
        """Test if experiment name conflict is automatically resolved when new name provided"""
        new_name = "test2"
        storage.create_experiment(parent_config)
        storage.create_experiment(child_config)
        child_config2 = copy.deepcopy(child_config)
        child_config2["version"] = 1
        conflicts = detect_conflicts(parent_config, child_config2)
        ExperimentBranchBuilder(conflicts, branch_to=new_name)

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 1

        conflict = conflicts.get()[0]

        assert conflict.resolution.new_name == new_name
        assert conflict.resolution.new_version == 1
        assert conflict.new_config["name"] == new_name
        assert conflict.new_config["version"] == 1
        assert conflict.is_resolved

    def test_bad_name_experiment(self, parent_config, child_config, monkeypatch):
        """Test if experiment name conflict is not resolved when invalid name is marked"""

        def _is_unique(self, *args, **kwargs):
            return False

        monkeypatch.setattr(
            ExperimentNameConflict.ExperimentNameResolution,
            "_name_is_unique",
            _is_unique,
        )

        conflicts = detect_conflicts(parent_config, child_config)
        ExperimentBranchBuilder(conflicts, branch_to="test2")

        assert len(conflicts.get()) == 1
        assert len(conflicts.get_resolved()) == 0

    def test_code_change(self, parent_config, changed_code_config):
        """Test if code conflict is resolved automatically"""
        change_type = evc.adapters.CodeChange.types[0]
        changed_code_config["code_change_type"] = change_type
        conflicts = detect_conflicts(parent_config, changed_code_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.CodeResolution)
        assert conflict.resolution.type == change_type

    def test_algo_change(self, parent_config, changed_algo_config):
        """Test if algorithm conflict is resolved automatically"""
        changed_algo_config["algorithm_change"] = True
        conflicts = detect_conflicts(parent_config, changed_algo_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.AlgorithmResolution)

    def test_orion_version_change(self, parent_config, changed_orion_version_config):
        """Test if orion version conflict is resolved automatically"""
        changed_orion_version_config["orion_version_change"] = True
        conflicts = detect_conflicts(parent_config, changed_orion_version_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.OrionVersionResolution)

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if user's script's config conflict is resolved automatically"""
        change_type = evc.adapters.ScriptConfigChange.types[0]
        changed_userconfig_config["config_change_type"] = change_type
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[1]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.ScriptConfigResolution)
        assert conflict.resolution.type == change_type

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if command line conflict is resolved automatically"""
        change_type = evc.adapters.CommandLineChange.types[0]
        changed_cli_config["cli_change_type"] = change_type
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        ExperimentBranchBuilder(conflicts, manual_resolution=True)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]

        assert conflict.is_resolved
        assert isinstance(conflict.resolution, conflict.CommandLineResolution)
        assert conflict.resolution.type == change_type


class TestAdapters(object):
    """Test creation of adapters"""

    def test_adapter_add_new(self, parent_config, cl_config):
        """Test if a DimensionAddition is created when solving a new conflict"""
        cl_config["metadata"]["user_args"] = ["-w_d~+normal(0,1)"]

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionAddition)

    def test_adapter_add_changed(self, parent_config, cl_config):
        """Test if a DimensionPriorChange is created when solving a new conflict"""
        cl_config["metadata"]["user_args"] = ["-y~+uniform(0,1)"]

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionPriorChange)

    def test_adapter_remove_missing(self, parent_config, cl_config):
        """Test if a DimensionDeletion is created when solving a new conflict"""
        cl_config["metadata"]["user_args"] = ["-z~-"]

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 2
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionDeletion)

    def test_adapter_rename_missing(self, parent_config, cl_config):
        """Test if a DimensionRenaming is created when solving a new conflict"""
        cl_config["metadata"]["user_args"] = ["-x~>w_d", "-w_d~+uniform(0,1)"]
        backward.populate_space(cl_config)

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 3
        assert len(adapters) == 1
        assert isinstance(adapters[0], evc.adapters.DimensionRenaming)

    def test_adapter_rename_different_prior(self, parent_config, cl_config):
        """Test if a DimensionRenaming is created when solving a new conflict"""
        cl_config["metadata"]["user_args"] = ["-x~>w_d", "-w_d~+normal(0,1)"]

        conflicts = detect_conflicts(parent_config, cl_config)
        branch_builder = ExperimentBranchBuilder(conflicts, manual_resolution=True)

        adapters = branch_builder.create_adapters().adapters

        assert len(conflicts.get_resolved()) == 4
        assert len(adapters) == 2
        assert isinstance(adapters[0], evc.adapters.DimensionRenaming)
        assert isinstance(adapters[1], evc.adapters.DimensionPriorChange)


class TestResolutionsConfig(object):
    """Test auto-resolution with specific types from orion.core.config.evc"""

    def test_cli_change(self, parent_config, changed_cli_config):
        """Test if giving a proper change-type solves the command line conflict"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        orion.core.config.evc.cli_change_type = "noeffect"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CommandLineConflict)
        assert conflict.resolution.type == "noeffect"
        orion.core.config.evc.cli_change_type = "break"

    def test_bad_cli_change(self, capsys, parent_config, changed_cli_config):
        """Test if giving an invalid change-type fails the the resolution"""
        conflicts = detect_conflicts(parent_config, changed_cli_config)
        orion.core.config.evc.cli_change_type = "bad-type"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1
        orion.core.config.evc.cli_change_type = "break"

    def test_code_change(self, parent_config, changed_code_config):
        """Test if giving a proper change-type solves the code conflict"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        orion.core.config.evc.code_change_type = "noeffect"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 2

        conflict = conflicts.get_resolved()[0]
        assert conflict.is_resolved
        assert isinstance(conflict, CodeConflict)
        assert conflict.resolution.type == "noeffect"
        orion.core.config.evc.code_change_type = "break"

    def test_bad_code_change(self, capsys, parent_config, changed_code_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_code_config)
        orion.core.config.evc.code_change_type = "bad-type"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 2
        assert len(conflicts.get_resolved()) == 1
        orion.core.config.evc.code_change_type = "break"

    def test_config_change(self, parent_config, changed_userconfig_config):
        """Test if giving a proper change-type solves the user script config conflict"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        orion.core.config.evc.config_change_type = "noeffect"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 4

        conflict = conflicts.get_resolved()[3]
        assert conflict.is_resolved
        assert isinstance(conflict, ScriptConfigConflict)
        assert conflict.resolution.type == "noeffect"

    def test_bad_config_change(self, capsys, parent_config, changed_userconfig_config):
        """Test if giving an invalid change-type prints error message and do nothing"""
        conflicts = detect_conflicts(parent_config, changed_userconfig_config)
        orion.core.config.evc.config_change_type = "bad-type"
        ExperimentBranchBuilder(conflicts)

        assert len(conflicts.get()) == 4
        assert len(conflicts.get_resolved()) == 3
        orion.core.config.evc.config_change_type = "break"



"""
Plotly backend for plotting methods
===================================

"""
import functools
from collections import Iterable

import numpy
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import orion.analysis
import orion.analysis.base
from orion.algo.space import Categorical, Fidelity
from orion.core.worker.transformer import build_required_space


def lpi(
    experiment,
    with_evc_tree=True,
    model="RandomForestRegressor",
    model_kwargs=None,
    n_points=20,
    n_runs=10,
    **kwargs,
):
    """Plotly implementation of `orion.plotting.lpi`"""
    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    if model_kwargs is None:
        model_kwargs = {}

    df = experiment.to_pandas(with_evc_tree=with_evc_tree)
    df = df.loc[df["status"] == "completed"]

    if df.empty:
        return go.Figure()

    df = orion.analysis.lpi(
        df,
        experiment.space,
        model=model,
        n_points=n_points,
        n_runs=n_runs,
        **model_kwargs,
    )

    fig = go.Figure(
        data=[
            go.Bar(
                x=df.index.tolist(),
                y=df["LPI"].tolist(),
                error_y=dict(type="data", array=df["STD"].tolist()),
            )
        ]
    )

    y_axis_label = "Local Parameter Importance (LPI)"
    fig.update_layout(
        title=f"LPI for experiment '{experiment.name}'",
        xaxis_title="Hyperparameters",
        yaxis_title=y_axis_label,
    )

    return fig


def parallel_coordinates(
    experiment, with_evc_tree=True, order=None, colorscale="YlOrRd", **kwargs
):
    """Plotly implementation of `orion.plotting.parallel_coordinates`"""

    def build_frame():
        """Builds the dataframe for the plot"""
        names = list(experiment.space.keys())

        df = experiment.to_pandas(with_evc_tree=with_evc_tree)
        df = df.loc[df["status"] == "completed"]

        if df.empty:
            return df

        df[names] = df[names].transform(
            functools.partial(_curate_params, space=experiment.space)
        )

        df = _flatten_dims(df, experiment.space)

        return df

    def infer_order(space, order):
        """Create order if not passed, otherwise verify it"""
        params = orion.analysis.base.flatten_params(space, order)
        if order is None:
            fidelity_dims = [
                dim for dim in experiment.space.values() if isinstance(dim, Fidelity)
            ]
            fidelity = fidelity_dims[0].name if fidelity_dims else None
            if fidelity in params:
                del params[params.index(fidelity)]
                params.insert(0, fidelity)

        return params

    def get_dimension(data, name, dim):
        dim_data = dict(label=name, values=data[name])
        if dim.type == "categorical":
            categories = dim.interval()
            dim_data["tickvals"] = list(range(len(categories)))
            dim_data["ticktext"] = categories
        else:
            dim_data["range"] = dim.interval()
        return dim_data

    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    df = build_frame()

    if df.empty:
        return go.Figure()

    trial = experiment.fetch_trials_by_status("completed")[0]

    flattened_space = build_required_space(
        experiment.space, shape_requirement="flattened"
    )

    dimensions = [
        get_dimension(df, name, flattened_space[name])
        for name in infer_order(experiment.space, order)
    ]

    objective_name = trial.objective.name

    objectives = df["objective"]
    omin = min(df["objective"])
    omax = max(df["objective"])

    dimensions.append(dict(label=objective_name, range=(omin, omax), values=objectives))

    fig = go.Figure(
        data=go.Parcoords(
            line=dict(
                color=objectives,
                colorscale=colorscale,
                showscale=True,
                cmin=omin,
                cmax=omax,
                colorbar=dict(title=objective_name),
            ),
            dimensions=dimensions,
        )
    )

    fig.update_layout(
        title=f"Parallel Coordinates Plot for experiment '{experiment.name}'"
    )

    return fig


def rankings(experiments, with_evc_tree=True, order_by="suggested", **kwargs):
    """Plotly implementation of `orion.plotting.rankings`"""

    def reformat_competitions(experiments):
        if isinstance(experiments, dict) and isinstance(
            next(iter(experiments.values())), Iterable
        ):
            competitions = []
            remaining = True
            i = 0
            n_competitions = len(next(iter(experiments.values())))
            for ith_competition in range(n_competitions):
                competition = {}
                for name in experiments.keys():
                    competition[name] = experiments[name][i]
                competitions.append(competition)
        elif isinstance(experiments, dict):
            competitions = experiments
        elif isinstance(experiments, Iterable) and not isinstance(experiments[0], dict):
            competitions = {
                f"{experiment.name}-v{experiment.version}": experiment
                for experiment in experiments
            }
        else:
            competitions = experiments

        return competitions

    def build_groups(competitions):

        if not isinstance(competitions, dict):
            rankings = []
            for competition in competitions:
                rankings.append(build_frame(competition))
            df = pd.concat(rankings)
            data_frames = orion.analysis.average(
                df, group_by=["order", "name"], key="rank", return_var=True
            )
        else:
            data_frames = build_frame(competitions)

        return data_frames

    def build_frame(competition):
        """Builds the dataframe for the plot"""

        frames = []
        for name, experiment in competition.items():
            df = experiment.to_pandas(with_evc_tree=with_evc_tree)
            df = df.loc[df["status"] == "completed"]
            df = df.sort_values(order_by)
            df = orion.analysis.regret(df)
            df["name"] = [name] * len(df)
            df["order"] = range(len(df))
            frames.append(df)

        df = pd.concat(frames)

        return orion.analysis.ranking(df)

    def get_objective_name(competition):
        """Infer name of objective based on trials of one experiment"""
        if not isinstance(competition, dict):
            return get_objective_name(competition[0])

        for experiment in competition.values():
            trials = experiment.fetch_trials_by_status("completed")
            if trials:
                return trials[0].objective.name
        return "objective"

    ORDER_KEYS = ["suggested", "reserved", "completed"]

    if not experiments:
        raise ValueError("Parameter 'experiment' is None")

    if order_by not in ORDER_KEYS:
        raise ValueError(f"Parameter 'order_by' is not one of {ORDER_KEYS}")

    competitions = reformat_competitions(experiments)
    df = build_groups(competitions)

    fig = go.Figure()

    if df.empty:
        return fig

    names = set(df["name"])
    for i, name in enumerate(sorted(names)):
        exp_data = df[df["name"] == name]
        if "rank_mean" in exp_data:
            y = exp_data["rank_mean"]
        else:
            y = exp_data["rank"]
        x = list(range(len(y)))
        fig.add_scatter(
            x=x,
            y=y,
            mode="lines",
            line=dict(color=px.colors.qualitative.G10[i]),
            name=name,
        )
        if "rank_var" in exp_data:
            dy = exp_data["rank_var"]
            fig.add_scatter(
                x=list(x) + list(x)[::-1],
                y=list(y - dy) + list(y + dy)[::-1],
                fill="toself",
                showlegend=False,
                line=dict(color=px.colors.qualitative.G10[i], width=0),
                name=name,
            )

    objective = get_objective_name(competitions)
    fig.update_layout(
        title=f"Average Rankings",
        xaxis_title=f"Trials ordered by {order_by} time",
        yaxis_title=f"Ranking based on {objective}",
        hovermode="x",
    )
    return fig


def partial_dependencies(
    experiment,
    with_evc_tree=True,
    params=None,
    smoothing=0.85,
    n_grid_points=10,
    n_samples=50,
    colorscale="Blues",
    model="RandomForestRegressor",
    model_kwargs=None,
    verbose_hover=True,
):
    """Plotly implementation of `orion.plotting.partial_dependencies`"""

    def build_data():
        """Builds the dataframe for the plot"""
        df = experiment.to_pandas(with_evc_tree=with_evc_tree)

        names = list(experiment.space.keys())
        df["params"] = df[names].apply(_format_hyperparameters, args=(names,), axis=1)

        df = df.loc[df["status"] == "completed"]
        data = orion.analysis.partial_dependency(
            df,
            experiment.space,
            params=params,
            model=model,
            n_grid_points=n_grid_points,
            n_samples=n_samples,
            **model_kwargs,
        )
        df = _flatten_dims(df, experiment.space)
        return (df, data)

    def _set_scale(figure, dims, x, y):
        for axis, dim in zip("xy", dims):
            if "reciprocal" in dim.prior_name or dim.type == "fidelity":
                getattr(figure, f"update_{axis}axes")(type="log", row=y, col=x)

    def _plot_marginalized_avg(data, x_name):
        return go.Scatter(
            x=data[0][x_name],
            y=data[1],
            mode="lines",
            name=None,
            showlegend=False,
            line=dict(
                color=px.colors.qualitative.D3[0],
            ),
        )

    def _plot_marginalized_std(data, x_name):
        return go.Scatter(
            x=list(data[0][x_name]) + list(data[0][x_name])[::-1],
            y=list(data[1] - data[2]) + list((data[1] + data[2]))[::-1],
            mode="lines",
            name=None,
            fill="toself",
            showlegend=False,
            line=dict(
                color=px.colors.qualitative.D3[0],
                width=0,
            ),
        )

    def _plot_contour(data, x_name, y_name):
        return go.Contour(
            x=data[0][x_name],
            y=data[0][y_name],
            z=data[1],
            connectgaps=True,
            # Share the same color range across contour plots
            coloraxis="coloraxis",
            line_smoothing=smoothing,
            # To show labels
            contours=dict(
                coloring="heatmap",
                showlabels=True,  # show labels on contours
                labelfont=dict(  # label font properties
                    size=12,
                    color="white",
                ),
            ),
        )

    def _plot_scatter(x, y, df):
        return go.Scatter(
            x=x,
            y=y,
            marker={
                "line": {"width": 0.5, "color": "Grey"},
                "color": "black",
                "size": 5,
            },
            mode="markers",
            opacity=0.5,
            showlegend=False,
            customdata=list(zip(df["id"], df["suggested"], df["params"])),
            hovertemplate=_template_trials(verbose_hover),
        )

    if model_kwargs is None:
        model_kwargs = {}

    df, data = build_data()

    if not data:
        return go.Figure()

    params = [
        param_names for param_names in data.keys() if isinstance(param_names, str)
    ]

    flattened_space = build_required_space(
        experiment.space,
        shape_requirement="flattened",
    )

    fig = make_subplots(
        rows=len(params),
        cols=len(params),
        shared_xaxes=True,
        shared_yaxes=False,
    )

    fig.update_layout(paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    cmin = float("inf")
    cmax = -float("inf")

    for x_i in range(len(params)):
        x_name = params[x_i]
        fig.add_trace(
            _plot_marginalized_avg(data[x_name], x_name),
            row=x_i + 1,
            col=x_i + 1,
        )
        fig.add_trace(
            _plot_marginalized_std(data[x_name], x_name),
            row=x_i + 1,
            col=x_i + 1,
        )
        fig.add_trace(
            _plot_scatter(df[x_name], df["objective"], df),
            row=x_i + 1,
            col=x_i + 1,
        )

        _set_scale(fig, [flattened_space[x_name]], x_i + 1, x_i + 1)

        fig.update_xaxes(title_text=x_name, row=len(params), col=x_i + 1)
        if x_i > 0:
            fig.update_yaxes(title_text=x_name, row=x_i + 1, col=1)
        else:
            fig.update_yaxes(title_text="Objective", row=x_i + 1, col=x_i + 1)

        for y_i in range(x_i + 1, len(params)):
            y_name = params[y_i]
            fig.add_trace(
                _plot_contour(
                    data[(x_name, y_name)],
                    x_name,
                    y_name,
                ),
                row=y_i + 1,
                col=x_i + 1,
            )
            fig.add_trace(
                _plot_scatter(df[x_name], df[y_name], df),
                row=y_i + 1,
                col=x_i + 1,
            )

            cmin = min(cmin, data[(x_name, y_name)][1].min())
            cmax = max(cmax, data[(x_name, y_name)][1].max())

            _set_scale(
                fig,
                [flattened_space[name] for name in [x_name, y_name]],
                x_i + 1,
                y_i + 1,
            )

    for x_i in range(len(params)):
        plot_id = len(params) * x_i + x_i + 1
        if plot_id > 1:
            key = f"yaxis{plot_id}_range"
        else:
            key = "yaxis_range"
        fig.update_layout(**{key: [cmin, cmax]})

    fig.update_layout(
        title=f"Partial dependencies for experiment '{experiment.name}'",
    )
    fig.layout.coloraxis.colorbar.title = "Objective"

    fig.update_layout(coloraxis=dict(colorscale=colorscale), showlegend=False)

    return fig


def regret(
    experiment, with_evc_tree=True, order_by="suggested", verbose_hover=True, **kwargs
):
    """Plotly implementation of `orion.plotting.regret`"""

    def build_frame():
        """Builds the dataframe for the plot"""
        df = experiment.to_pandas(with_evc_tree=with_evc_tree)

        names = list(experiment.space.keys())
        df["params"] = df[names].apply(_format_hyperparameters, args=(names,), axis=1)

        df = df.loc[df["status"] == "completed"]
        df = df.sort_values(order_by)
        df = orion.analysis.regret(df)
        return df

    ORDER_KEYS = ["suggested", "reserved", "completed"]

    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    if order_by not in ORDER_KEYS:
        raise ValueError(f"Parameter 'order_by' is not one of {ORDER_KEYS}")

    df = build_frame()

    fig = go.Figure()

    if df.empty:
        return fig

    trial = experiment.fetch_trials_by_status("completed")[0]

    fig.add_scatter(
        y=df["objective"],
        mode="markers",
        name="trials",
        customdata=list(zip(df["id"], df[order_by], df["params"])),
        hovertemplate=_template_trials(verbose_hover),
    )
    fig.add_scatter(
        y=df["best"],
        mode="lines",
        name="best-to-date",
        customdata=list(zip(df["best_id"], df["best"])),
        hovertemplate=_template_best(),
    )

    if trial is None:
        y_axis_label = "Objective unknown"
    else:
        y_axis_label = f"{trial.objective.type.capitalize()} '{trial.objective.name}'"

    fig.update_layout(
        title=f"Regret for experiment '{experiment.name}'",
        xaxis_title=f"Trials ordered by {order_by} time",
        yaxis_title=y_axis_label,
    )

    return fig


def regrets(experiments, with_evc_tree=True, order_by="suggested", **kwargs):
    """Plotly implementation of `orion.plotting.regrets`"""

    compute_average = bool(
        isinstance(experiments, dict)
        and isinstance(next(iter(experiments.values())), Iterable)
    )

    def build_groups():
        """Build dataframes for groups of experiments"""
        # TODO move this
        if compute_average:
            data_frames = dict()
            for name, group in experiments.items():
                df = orion.analysis.average(build_frame(group), return_var=True)
                df["name"] = [name] * len(df)
                data_frames[name] = df
            data_frames = pd.concat(data_frames)
        elif isinstance(experiments, dict):
            data_frames = build_frame(experiments.values(), list(experiments.keys()))
        else:
            data_frames = build_frame(experiments)

        return data_frames

    def build_frame(experiments, names=None):
        """Builds the dataframe for the plot"""
        frames = []
        for i, experiment in enumerate(experiments):
            df = experiment.to_pandas(with_evc_tree=with_evc_tree)
            df = df.loc[df["status"] == "completed"]
            df = df.sort_values(order_by)
            df = orion.analysis.regret(df)
            df["name"] = [
                names[i] if names else f"{experiment.name}-v{experiment.version}"
            ] * len(df)
            df["order"] = range(len(df))
            frames.append(df)

        return pd.concat(frames)

    def get_objective_name(experiments):
        """Infer name of objective based on trials of one experiment"""
        if compute_average and isinstance(experiments, dict):
            return get_objective_name(sum(map(list, experiments.values()), []))

        if isinstance(experiments, dict):
            experiments = experiments.values()

        for experiment in experiments:
            trials = experiment.fetch_trials_by_status("completed")
            if trials:
                return trials[0].objective.name
        return "objective"

    ORDER_KEYS = ["suggested", "reserved", "completed"]

    if not experiments:
        raise ValueError("Parameter 'experiment' is None")

    if order_by not in ORDER_KEYS:
        raise ValueError(f"Parameter 'order_by' is not one of {ORDER_KEYS}")

    df = build_groups()

    fig = go.Figure()

    if df.empty:
        return fig

    names = set(df["name"])
    for i, name in enumerate(sorted(names)):
        exp_data = df[df["name"] == name]
        if "best_mean" in exp_data:
            y = exp_data["best_mean"]
        else:
            y = exp_data["best"]
        x = list(range(len(y)))
        fig.add_scatter(
            x=x,
            y=y,
            mode="lines",
            line=dict(color=px.colors.qualitative.G10[i]),
            name=name,
        )
        if "best_var" in exp_data:
            dy = exp_data["best_var"]
            fig.add_scatter(
                x=list(x) + list(x)[::-1],
                y=list(y - dy) + list(y + dy)[::-1],
                fill="toself",
                showlegend=False,
                line=dict(color=px.colors.qualitative.G10[i]),
                name=name,
            )

    fig.update_layout(
        title=f"Average Regret",
        xaxis_title=f"Trials ordered by {order_by} time",
        yaxis_title=get_objective_name(experiments),
        hovermode="x",
    )

    return fig


def _format_value(value):
    """
    Hyperparameter can have many types, sometimes they can even be lists.
    If one of the value is a float, it has to be compact.
    """
    if isinstance(value, str):
        return value

    try:
        return f"[{','.join(_format_value(x) for x in value)}]"

    except TypeError:
        if isinstance(value, float):
            return f"{value:.5G}"
        else:
            return value


def _format_hyperparameters(hyperparameters, names):
    result = ""

    for name, value in zip(names, hyperparameters):
        x = f"<br>  {name}: {_format_value(value)}"
        result += x

    return result


def _template_trials(verbose_hover):
    template = (
        "<b>ID: %{customdata[0]}</b><br>"
        "value: %{y}<br>"
        "time: %{customdata[1]|%Y-%m-%d %H:%M:%S}<br>"
    )

    if verbose_hover:
        template += "parameters: %{customdata[2]}"

    template += "<extra></extra>"

    return template


def _template_best():
    return (
        "<b>Best ID: %{customdata[0]}</b><br>"
        "value: %{customdata[1]}"
        "<extra></extra>"
    )


def _curate_params(data, space):
    dim = space[data.name]
    if isinstance(dim, Categorical):
        data = numpy.array(data.tolist())  # To unpack lists if dim shape > 1
        shape = data.shape
        assert len(shape) <= 2
        idx = numpy.argmax(data.reshape(-1, 1) == numpy.array(dim.categories), axis=1)
        idx = idx.reshape(shape)
        if len(shape) > 1:
            return [list(idx[i]) for i in range(shape[0])]
        return idx
    return data


def _flatten_dims(data, space):
    for key, dim in space.items():
        if dim.shape:
            assert len(dim.shape) == 1

            # expand df.tags into its own dataframe
            values = data[key].apply(pd.Series)

            # rename each hp
            values = values.rename(columns=lambda x: f"{key}[{x}]")

            data = pd.concat([data[:], values[:]], axis=1)

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.resolve_config`."""

import hashlib
import logging
import os
import shutil
import socket

import git
import pytest

import orion.core
import orion.core.io.resolve_config as resolve_config


@pytest.fixture
def force_is_exe(monkeypatch):
    """Mock resolve_config to recognize any string as an executable script."""

    def is_exe(path):
        return True

    monkeypatch.setattr(resolve_config, "is_exe", is_exe)


def test_fetch_env_vars():
    """Verify env vars are fetched properly"""
    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {"database": {}}

    db_name = "orion_test"

    os.environ["ORION_DB_NAME"] = db_name

    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {"database": {"name": "orion_test"}}

    db_type = "MongoDB"
    os.environ["ORION_DB_TYPE"] = db_type

    env_vars_config = resolve_config.fetch_env_vars()
    assert env_vars_config == {"database": {"name": db_name, "type": db_type}}


@pytest.mark.usefixtures("version_XYZ")
def test_fetch_metadata_orion_version():
    """Verify orion version"""
    metadata = resolve_config.fetch_metadata()
    assert metadata["orion_version"] == "XYZ"


def test_fetch_metadata_executable_users_script(script_path):
    """Verify executable user script with absolute path"""
    metadata = resolve_config.fetch_metadata(user_args=[script_path])
    assert metadata["user_script"] == os.path.abspath(script_path)


def test_fetch_metadata_non_executable_users_script():
    """Verify executable user script keeps given path"""
    script_path = "tests/functional/demo/orion_config.yaml"
    metadata = resolve_config.fetch_metadata(user_args=[script_path])
    assert metadata["user_script"] == script_path


def test_fetch_metadata_python_users_script(script_path):
    """Verify user script is correctly infered if called with python"""
    metadata = resolve_config.fetch_metadata(
        user_args=["python", script_path, "some", "args"]
    )
    assert metadata["user_script"] == script_path


def test_fetch_metadata_not_existed_path():
    """Verfiy the raise of error when user_script path does not exist"""
    path = "dummy/path"
    with pytest.raises(OSError) as exc_info:
        resolve_config.fetch_metadata(user_args=[path])
    assert "The path specified for the script does not exist" in str(exc_info.value)


@pytest.mark.usefixtures()
def test_fetch_metadata_user_args(script_path):
    """Verify user args"""
    user_args = [os.path.abspath(script_path)] + list(map(str, range(10)))
    metadata = resolve_config.fetch_metadata(user_args=user_args)
    assert metadata["user_script"] == user_args[0]
    assert metadata["user_args"] == user_args


@pytest.mark.usefixtures("with_user_tsirif")
def test_fetch_metadata_user_tsirif():
    """Verify user name"""
    metadata = resolve_config.fetch_metadata()
    assert metadata["user"] == "tsirif"


def test_fetch_metadata():
    """Verify no additional data is stored in metadata"""
    metadata = resolve_config.fetch_metadata()
    len(metadata) == 4


def test_fetch_config_from_cmdargs():
    """Verify fetch_config returns empty dict on no config file path"""
    cmdargs = {
        "name": "test",
        "user": "me",
        "version": 1,
        "config": None,
        "exp_max_trials": "exp_max_trials",
        "worker_trials": "worker_trials",
        "exp_max_broken": "exp_max_broken",
        "working_dir": "working_dir",
        "pool_size": "pool_size",
        "max_trials": "max_trials",
        "heartbeat": "heartbeat",
        "worker_max_trials": "worker_max_trials",
        "worker_max_broken": "worker_max_broken",
        "max_idle_time": "max_idle_time",
        "interrupt_signal_code": "interrupt_signal_code",
        "user_script_config": "user_script_config",
        "manual_resolution": "manual_resolution",
        "non_monitored_arguments": "non_monitored_arguments",
        "ignore_code_changes": "ignore_code_changes",
        "auto_resolution": "auto_resolution",
        "branch_from": "branch_from",
        "algorithm_change": "algorithm_change",
        "code_change_type": "code_change_type",
        "cli_change_type": "cli_change_type",
        "branch_to": "branch_to",
        "config_change_type": "config_change_type",
    }

    config = resolve_config.fetch_config_from_cmdargs(cmdargs)

    assert config.pop("config", None) is None

    exp_config = config.pop("experiment")
    assert exp_config.pop("name") == "test"
    assert exp_config.pop("version") == 1
    assert exp_config.pop("user") == "me"
    assert exp_config.pop("max_trials") == "exp_max_trials"
    assert exp_config.pop("max_broken") == "exp_max_broken"
    assert exp_config.pop("working_dir") == "working_dir"
    assert exp_config.pop("pool_size") == "pool_size"

    assert exp_config == {}

    worker_config = config.pop("worker")
    assert worker_config.pop("heartbeat") == "heartbeat"
    assert worker_config.pop("max_trials") == "worker_max_trials"
    assert worker_config.pop("max_broken") == "worker_max_broken"
    assert worker_config.pop("max_idle_time") == "max_idle_time"
    assert worker_config.pop("interrupt_signal_code") == "interrupt_signal_code"
    assert worker_config.pop("user_script_config") == "user_script_config"

    assert worker_config == {}

    evc_config = config.pop("evc")
    assert evc_config.pop("manual_resolution") == "manual_resolution"
    assert evc_config.pop("non_monitored_arguments") == "non_monitored_arguments"
    assert evc_config.pop("ignore_code_changes") == "ignore_code_changes"
    assert evc_config.pop("auto_resolution") == "auto_resolution"
    assert evc_config.pop("branch_from") == "branch_from"
    assert evc_config.pop("algorithm_change") == "algorithm_change"
    assert evc_config.pop("code_change_type") == "code_change_type"
    assert evc_config.pop("cli_change_type") == "cli_change_type"
    assert evc_config.pop("branch_to") == "branch_to"
    assert evc_config.pop("config_change_type") == "config_change_type"

    assert evc_config == {}

    assert config == {}


@pytest.mark.parametrize(
    "argument",
    ["config", "user", "user_args", "name", "version", "branch_from", "branch_to"],
)
def test_fetch_config_from_cmdargs_no_empty(argument):
    """Verify fetch_config returns only defined arguments."""
    config = resolve_config.fetch_config_from_cmdargs({})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: None})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: False})
    assert config == {}

    config = resolve_config.fetch_config_from_cmdargs({argument: 1})

    if argument in ["name", "user", "version"]:
        assert config == {"experiment": {argument: 1}}
    elif argument in ["branch_from", "branch_to"]:
        assert config == {"evc": {argument: 1}}
    else:
        assert config == {argument: 1}


def test_fetch_config_no_hit():
    """Verify fetch_config returns empty dict on no config file path"""
    config = resolve_config.fetch_config({"config": ""})
    assert config == {}


def test_fetch_config(config_file):
    """Verify fetch_config returns valid dictionnary"""
    config = resolve_config.fetch_config({"config": config_file})

    assert config.pop("storage") == {
        "database": {
            "host": "mongodb://user:pass@localhost",
            "name": "orion_test",
            "type": "mongodb",
        }
    }

    assert config.pop("experiment") == {
        "max_trials": 100,
        "max_broken": 5,
        "name": "voila_voici",
        "pool_size": 1,
        "algorithms": "random",
        "strategy": "NoParallelStrategy",
    }

    assert config == {}


def test_fetch_config_global_local_coherence(monkeypatch, config_file):
    """Verify fetch_config parses local config according to global config structure."""

    def mocked_config(file_object):
        return orion.core.config.to_dict()

    monkeypatch.setattr("yaml.safe_load", mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    # Test storage subconfig
    storage_config = config.pop("storage")
    database_config = storage_config.pop("database")
    assert storage_config.pop("type") == orion.core.config.storage.type
    assert storage_config == {}

    assert database_config.pop("host") == orion.core.config.storage.database.host
    assert database_config.pop("name") == orion.core.config.storage.database.name
    assert database_config.pop("port") == orion.core.config.storage.database.port
    assert database_config.pop("type") == orion.core.config.storage.database.type

    assert database_config == {}

    # Test experiment subconfig
    exp_config = config.pop("experiment")
    assert exp_config.pop("max_trials") == orion.core.config.experiment.max_trials
    assert exp_config.pop("max_broken") == orion.core.config.experiment.max_broken
    assert exp_config.pop("working_dir") == orion.core.config.experiment.working_dir
    assert exp_config.pop("pool_size") == orion.core.config.experiment.pool_size
    assert exp_config.pop("algorithms") == orion.core.config.experiment.algorithms
    assert exp_config.pop("strategy") == orion.core.config.experiment.strategy

    assert exp_config == {}

    # Test worker subconfig
    worker_config = config.pop("worker")
    assert worker_config.pop("n_workers") == orion.core.config.worker.n_workers
    assert worker_config.pop("executor") == orion.core.config.worker.executor
    assert (
        worker_config.pop("executor_configuration")
        == orion.core.config.worker.executor_configuration
    )
    assert worker_config.pop("heartbeat") == orion.core.config.worker.heartbeat
    assert worker_config.pop("max_trials") == orion.core.config.worker.max_trials
    assert worker_config.pop("max_broken") == orion.core.config.worker.max_broken
    assert worker_config.pop("max_idle_time") == orion.core.config.worker.max_idle_time
    assert (
        worker_config.pop("interrupt_signal_code")
        == orion.core.config.worker.interrupt_signal_code
    )
    assert (
        worker_config.pop("user_script_config")
        == orion.core.config.worker.user_script_config
    )

    assert worker_config == {}

    # Test evc subconfig
    evc_config = config.pop("evc")
    assert evc_config.pop("auto_resolution") == orion.core.config.evc.auto_resolution
    assert (
        evc_config.pop("manual_resolution") == orion.core.config.evc.manual_resolution
    )
    assert (
        evc_config.pop("non_monitored_arguments")
        == orion.core.config.evc.non_monitored_arguments
    )
    assert (
        evc_config.pop("ignore_code_changes")
        == orion.core.config.evc.ignore_code_changes
    )
    assert evc_config.pop("algorithm_change") == orion.core.config.evc.algorithm_change
    assert (
        evc_config.pop("orion_version_change")
        == orion.core.config.evc.orion_version_change
    )
    assert evc_config.pop("code_change_type") == orion.core.config.evc.code_change_type
    assert evc_config.pop("cli_change_type") == orion.core.config.evc.cli_change_type
    assert (
        evc_config.pop("config_change_type") == orion.core.config.evc.config_change_type
    )

    assert evc_config == {}

    # Confirm that all fields were tested.
    assert config == {}


def test_fetch_config_dash(monkeypatch, config_file):
    """Verify fetch_config supports dash."""

    def mocked_config(file_object):
        return {"experiment": {"max-broken": 10, "algorithms": {"dont-change": "me"}}}

    monkeypatch.setattr("yaml.safe_load", mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {
        "experiment": {"max_broken": 10, "algorithms": {"dont-change": "me"}}
    }


def test_fetch_config_underscore(monkeypatch, config_file):
    """Verify fetch_config supports underscore as well."""

    def mocked_config(file_object):
        return {"experiment": {"max_broken": 10, "algorithms": {"dont-change": "me"}}}

    monkeypatch.setattr("yaml.safe_load", mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {
        "experiment": {"max_broken": 10, "algorithms": {"dont-change": "me"}}
    }


def test_fetch_config_deprecated_max_trials(monkeypatch, config_file):
    """Verify fetch_config will overwrite deprecated value if also properly defined."""

    def mocked_config(file_object):
        return {"experiment": {"max_trials": 10}, "max_trials": 20}

    monkeypatch.setattr("yaml.safe_load", mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {"experiment": {"max_trials": 10}}


def test_fetch_config_deprecate_tricky_names(monkeypatch, config_file):
    """Verify fetch_config assigns values properly for the similar names."""

    def mocked_config(file_object):
        return {
            "experiment": {"worker_trials": "should_be_ignored"},
            "max_trials": "exp_max_trials",
            "max_broken": "exp_max_broken",
            "worker_trials": "worker_max_trials",
            "name": "exp_name",
        }

    monkeypatch.setattr("yaml.safe_load", mocked_config)

    config = resolve_config.fetch_config({"config": config_file})

    assert config == {
        "experiment": {
            "name": "exp_name",
            "max_trials": "exp_max_trials",
            "max_broken": "exp_max_broken",
        },
        "worker": {"max_trials": "worker_max_trials"},
    }


def test_merge_configs_update_two():
    """Ensure update on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"a": 3}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 3, "b": 2}


def test_merge_configs_update_three():
    """Ensure two updates on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"a": 3}
    c = {"b": 4}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {"a": 3, "b": 4}


def test_merge_configs_update_four():
    """Ensure three updates on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"a": 3}
    c = {"b": 4}
    d = {"a": 5, "b": 6}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {"a": 5, "b": 6}


def test_merge_configs_extend_two():
    """Ensure extension on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"c": 3}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": 2, "c": 3}


def test_merge_configs_extend_three():
    """Ensure two extensions on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"c": 3}
    c = {"d": 4}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {"a": 1, "b": 2, "c": 3, "d": 4}


def test_merge_configs_extend_four():
    """Ensure three extensions on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"c": 3}
    c = {"d": 4}
    d = {"e": 5}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}


def test_merge_configs_update_extend_two():
    """Ensure update and extension on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"b": 3, "c": 4}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": 3, "c": 4}


def test_merge_configs_update_extend_three():
    """Ensure two updates and extensions on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"b": 3, "c": 4}
    c = {"a": 5, "d": 6}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {"a": 5, "b": 3, "c": 4, "d": 6}


def test_merge_configs_update_extend_four():
    """Ensure three updates and extensions on first level is fine"""
    a = {"a": 1, "b": 2}
    b = {"b": 3, "c": 4}
    c = {"a": 5, "d": 6}
    d = {"d": 7, "e": 8}

    m = resolve_config.merge_configs(a, b, c, d)

    assert m == {"a": 5, "b": 3, "c": 4, "d": 7, "e": 8}


def test_merge_sub_configs_update_two():
    """Ensure updating to second level is fine"""
    a = {"a": 1, "b": 2}
    b = {"b": {"c": 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": {"c": 3}}


def test_merge_sub_configs_sub_update_two():
    """Ensure updating on second level is fine"""
    a = {"a": 1, "b": {"c": 2}}
    b = {"b": {"c": 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": {"c": 3}}

    a = {"a": 1, "b": {"c": 2, "d": 3}}
    b = {"b": {"c": 4}}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": {"c": 4, "d": 3}}


def test_merge_sub_configs_sub_extend_two():
    """Ensure updating to third level from second level is fine"""
    a = {"a": 1, "b": {"c": 2}}
    b = {"d": {"e": 3}}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": {"c": 2}, "d": {"e": 3}}

    a = {"a": 1, "b": {"c": 2, "d": 3}}
    b = {"b": {"e": {"f": 4}}}

    m = resolve_config.merge_configs(a, b)

    assert m == {"a": 1, "b": {"c": 2, "d": 3, "e": {"f": 4}}}


def test_merge_sub_configs_update_three():
    """Ensure updating twice to third level from second level is fine"""
    a = {"a": 1, "b": {"c": 2}}
    b = {"b": {"c": 3}}
    c = {"b": {"c": {"d": 4}}}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {"a": 1, "b": {"c": {"d": 4}}}

    a = {"a": 1, "b": {"c": 2, "d": 3}}
    b = {"b": {"c": 4}}
    c = {"b": {"c": {"e": 5}}}

    m = resolve_config.merge_configs(a, b, c)

    assert m == {"a": 1, "b": {"c": {"e": 5}, "d": 3}}


def test_merge_matching_type_configs():
    """Test that configs with matching type are merged properly"""
    a = {"a": 1, "b": {"c": 2, "t": "match"}}
    b = {"b": {"c": 3, "d": 4, "t": "match"}}
    c = {"b": {"c": {"d": 4}, "e": 5, "t": "match"}}

    m = resolve_config.merge_configs(a, b, c, differentiators=["t"])

    assert m == {"a": 1, "b": {"c": {"d": 4}, "d": 4, "e": 5, "t": "match"}}


def test_merge_diff_type_configs():
    """Test that configs with diff type are not merged"""
    a = {"a": 1, "b": {"c": 2, "t": 1}}
    b = {"a": 1, "b": {"c": 3, "d": 4, "t": 2}}
    c = {"b": {"c": {"d": 4}, "e": 5, "t": 3}}

    m = resolve_config.merge_configs(a, b, differentiators=["t"])

    assert m == {"a": 1, "b": {"c": 3, "d": 4, "t": 2}}

    m = resolve_config.merge_configs(b, c, differentiators=["t"])
    assert m == {"a": 1, "b": {"c": {"d": 4}, "e": 5, "t": 3}}

    assert resolve_config.merge_configs(
        b, c, differentiators=["t"]
    ) == resolve_config.merge_configs(a, b, c, differentiators=["t"])


def test_merge_diff_type_sub_configs():
    """Test that configs with nested diff type are not merged"""
    a = {"a": 1, "b": {"c": 2, "t": 1, "d": {"t": 2, "e": 3}}}
    b = {"b": {"a": 3, "t": 1, "d": {"t": 2, "f": 4}}}

    m = resolve_config.merge_configs(a, b, differentiators=["t"])
    assert m == {"a": 1, "b": {"a": 3, "c": 2, "t": 1, "d": {"t": 2, "e": 3, "f": 4}}}

    c = {"b": {"a": 3, "t": 1, "d": {"t": 3, "f": 4}}}

    m = resolve_config.merge_configs(a, c, differentiators=["t"])
    assert m == {"a": 1, "b": {"a": 3, "c": 2, "t": 1, "d": {"t": 3, "f": 4}}}

    d = {"b": {"a": 3, "t": 2, "d": {"t": 2, "f": 4}}}

    m = resolve_config.merge_configs(a, d, differentiators=["t"])
    assert m == {"a": 1, "b": {"a": 3, "t": 2, "d": {"t": 2, "f": 4}}}


@pytest.fixture
def repo():
    """Create a dummy repo for the tests."""
    os.chdir("../")
    os.makedirs("dummy_orion")
    os.chdir("dummy_orion")
    repo = git.Repo.init(".")
    with open("README.md", "w+") as f:
        f.write("dummy content")
    repo.git.add("README.md")
    repo.index.commit("initial commit")
    repo.create_head("master")
    repo.git.checkout("master")
    yield repo
    os.chdir("../")
    shutil.rmtree("dummy_orion")
    os.chdir("orion")


def test_infer_versioning_metadata_on_clean_repo(repo):
    """
    Test how `infer_versioning_metadata` fills its different fields
    when the user's repo is clean:
    `is_dirty`, `active_branch` and `diff_sha`.
    """
    vcs = resolve_config.infer_versioning_metadata(".git")
    assert not vcs["is_dirty"]
    assert vcs["active_branch"] == "master"
    # the diff should be empty so the diff_sha should be equal to the diff sha of an empty string
    assert vcs["diff_sha"] == hashlib.sha256("".encode("utf-8")).hexdigest()


def test_infer_versioning_metadata_on_dirty_repo(repo):
    """
    Test how `infer_versioning_metadata` fills its different fields
    when the uers's repo is dirty:
    `is_dirty`, `HEAD_sha`, `active_branch` and `diff_sha`.
    """
    existing_metadata = {}
    existing_metadata["user_script"] = ".git"
    vcs = resolve_config.infer_versioning_metadata(".git")
    repo.create_head("feature")
    repo.git.checkout("feature")
    with open("README.md", "w+") as f:
        f.write("dummy dummy content")
    vcs = resolve_config.infer_versioning_metadata(".git")
    assert vcs["is_dirty"]
    assert vcs["active_branch"] == "feature"
    assert vcs["diff_sha"] != hashlib.sha256("".encode("utf-8")).hexdigest()
    repo.git.add("README.md")
    commit = repo.index.commit("Added dummy_file")
    vcs = resolve_config.infer_versioning_metadata(".git")
    assert not vcs["is_dirty"]
    assert vcs["HEAD_sha"] == commit.hexsha
    assert vcs["diff_sha"] == hashlib.sha256("".encode("utf-8")).hexdigest()


def test_fetch_user_repo_on_non_repo(caplog):
    """Test if `fetch_user_repo` logs a warning when user's script is not a git repo."""
    with caplog.at_level(logging.WARNING):
        resolve_config.fetch_user_repo(".")
    messages = [rec.message for rec in caplog.records]
    assert len(messages) == 1
    assert "Script {} is not in a git repository".format(os.getcwd()) in messages[0]


def test_infer_versioning_metadata_on_detached_head(repo):
    """Test in the case of a detached head."""
    with open("README.md", "w+") as f:
        f.write("dummy contentt")
    repo.git.add("README.md")
    repo.index.commit("2nd commit")
    existing_metadata = {}
    existing_metadata["user_script"] = ".git"
    repo.head.reference = repo.commit("HEAD~1")
    assert repo.head.is_detached
    vcs = resolve_config.infer_versioning_metadata(".git")
    assert vcs["active_branch"] is None



"""
Tools to compute Partial Dependency
===================================
"""
import copy
import itertools

import numpy
import pandas

from orion.analysis.base import flatten_numpy, flatten_params, to_numpy, train_regressor
from orion.core.worker.transformer import build_required_space


def partial_dependency(
    trials,
    space,
    params=None,
    model="RandomForestRegressor",
    n_grid_points=10,
    n_samples=50,
    **kwargs
):
    """
    Calculates the partial dependency of parameters in a collection of
    :class:`orion.core.worker.trial.Trial`.

    Parameters
    ----------
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    space: Space object
        A space object from an experiment.

    params: list of str, optional
        The parameters to include in the computation. All parameters are included by default.

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

    n_grid_points: int
        Number of points in the grid to compute partial dependency. Default is 10.

    n_samples: int
        Number of samples to randomly generate the grid used to compute the partial dependency.
        Default is 50.

    **kwargs
        Arguments for the regressor model.

    Returns
    -------
    dict
        Dictionary of DataFrames. Each combination of parameters as keys (dim1.name, dim2.name)
        and for each parameters individually (dim1.name). Columns are
        (dim1.name, dim2.name, objective) or (dim1.name, objective).

    """
    params = flatten_params(space, params)

    flattened_space = build_required_space(
        space,
        dist_requirement="linear",
        type_requirement="numerical",
        shape_requirement="flattened",
    )

    if trials.empty or trials.shape[0] == 0:
        return {}

    data = to_numpy(trials, space)
    data = flatten_numpy(data, flattened_space)
    model = train_regressor(model, data, **kwargs)

    data = flattened_space.sample(n_samples)
    data = pandas.DataFrame(data, columns=flattened_space.keys())

    partial_dependencies = dict()
    for x_i, x_name in enumerate(params):
        grid, averages, stds = partial_dependency_grid(
            flattened_space, model, [x_name], data, n_grid_points
        )
        grid = reverse(flattened_space, grid)
        partial_dependencies[x_name] = (grid, averages, stds)
        for y_i in range(x_i + 1, len(params)):
            y_name = params[y_i]
            grid, averages, stds = partial_dependency_grid(
                flattened_space, model, [x_name, y_name], data, n_grid_points
            )
            grid = reverse(flattened_space, grid)
            partial_dependencies[(x_name, y_name)] = (grid, averages, stds)

    return partial_dependencies


def reverse(transformed_space, grid):
    """Reverse transformations on the grid to bring back to original space"""
    for param in grid.keys():
        transformed_dim = transformed_space[param].original_dimension
        param_grid = []
        for value in grid[param]:
            param_grid.append(transformed_dim.reverse(value))
        grid[param] = param_grid
    return grid


def make_grid(dim, n_points):
    """Build a grid of n_points for a dim"""
    if dim.prior_name == "choices":
        low, high = dim.interval()
        return numpy.arange(low, high + 1)

    return numpy.linspace(*dim.interval(), num=n_points)


def partial_dependency_grid(space, model, params, samples, n_points=40):
    """Compute the dependency grid for a given set of params (1 or 2)"""

    samples = copy.deepcopy(samples)

    grids = {}
    for name in params:
        grids[name] = make_grid(space[name], n_points)

    lengths = [len(grids[name]) for name in params]
    averages = numpy.zeros(lengths)
    stds = numpy.zeros(lengths)

    indexed_combinations = zip(
        itertools.product(*(list(range(length)) for length in lengths)),
        itertools.product(*grids.values()),
    )
    for z_idx, combination in indexed_combinations:
        for i, name in enumerate(params):
            samples[name] = combination[i]

        predictions = model.predict(samples.to_numpy())
        averages[z_idx] = numpy.mean(predictions)
        stds[z_idx] = numpy.std(predictions)

    return grids, averages.T, stds.T



# -*- coding: utf-8 -*-
"""
Different mutate functions: large-scale evolution of image classifiers
======================================================================

Implement evolution to exploit configurations with fixed resource efficiently

"""
import logging

import numpy as np

logger = logging.getLogger(__name__)


def default_mutate(search_space, rng, old_value, **kwargs):
    """Get a default mutate function"""
    multiply_factor = kwargs.pop("multiply_factor", 3.0)
    add_factor = kwargs.pop("add_factor", 1)
    volatility = kwargs.pop("volatility", 0.001)
    if search_space.type == "real":
        lower_bound, upper_bound = search_space.interval()
        factors = (
            1.0 / multiply_factor
            + (multiply_factor - 1.0 / multiply_factor) * rng.random()
        )
        if lower_bound <= old_value * factors <= upper_bound:
            new_value = old_value * factors
        elif lower_bound > old_value * factors:
            new_value = lower_bound + volatility * rng.random()
        else:
            new_value = upper_bound - volatility * rng.random()
    elif search_space.type == "integer":
        print(search_space)
        lower_bound, upper_bound = search_space.interval()
        factors = int(add_factor * (2 * rng.randint(2) - 1))
        if lower_bound <= old_value + factors <= upper_bound:
            new_value = int(old_value) + factors
        elif lower_bound > old_value + factors:
            new_value = int(lower_bound)
        else:
            new_value = int(upper_bound)
    elif search_space.type == "categorical":
        # TODO: This ignores the probabilities passed to search space.
        #       The mutation function should work directly at the search space level
        #       instead of separately on each dimensions. This would make it possible
        #       to sample properly the categorical dimensions.
        new_value = rng.choice(search_space.interval())
    else:
        print(search_space.type)
        new_value = old_value
    return new_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the info command."""
import orion.core.cli
import orion.core.io.resolve_config


def test_info_no_hit(one_experiment, capsys):
    """Test info if no experiment with given name."""
    returncode = orion.core.cli.main(["info", "--name", "i do not exist"])

    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured.startswith("Error: No experiment with given name 'i do not exist'")


def test_info_hit(one_experiment, capsys):
    """Test info if existing experiment."""
    orion.core.cli.main(["info", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    assert "--x~uniform(0,1)" in captured


def test_info_broken(broken_refers, capsys):
    """Test info if experiment.refers is missing."""
    orion.core.cli.main(["info", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    assert "--x~uniform(0,1)" in captured


def test_info_no_branching(one_experiment_changed_vcs, capsys):
    """Test info if config file is different

    Version should not increase!
    """
    orion.core.cli.main(["info", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    assert "\nversion: 1\n" in captured


def test_info_python_api(with_experiment_using_python_api, capsys):
    """Test info if config built using python api"""
    orion.core.cli.main(["info", "--name", "from-python-api"])

    captured = capsys.readouterr().out

    assert "from-python-api" in captured
    assert "Commandline" not in captured


def test_missing_conf_file(with_experiment_missing_conf_file, capsys):
    """Test info can handle experiments when the user script config file is missing"""
    orion.core.cli.main(["info", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    assert "--x~uniform(0,1)" in captured


def test_info_cmdline_api(one_experiment, capsys):
    """Test info if config built using cmdline api"""
    orion.core.cli.main(["info", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    assert "test_single_exp" in captured
    assert "Commandline" in captured


def test_no_args(capsys):
    """Try to run the command without any arguments"""
    returncode = orion.core.cli.main(["info"])
    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured == "Error: No name provided for the experiment.\n"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Script that will always interrupt trials."""
import argparse

from orion.client import (  # noqa: F401
    interrupt_trial,
    report_bad_trial,
    report_objective,
)


def no_report():
    """Do not report any result"""
    return


def execute():
    """Execute a simple pipeline as an example."""
    parser = argparse.ArgumentParser()
    parser.add_argument("fct", type=str)
    parser.add_argument("--name", type=str)
    parser.add_argument("--objective", type=str)
    parser.add_argument("--data", type=str)
    parser.add_argument("-x", type=float)

    inputs = parser.parse_args()

    kwargs = {}

    # Maybe its a float, maybe user made a mistake and report objective='name'
    try:
        inputs.objective = float(inputs.objective)
    except (ValueError, TypeError):
        pass

    for key, value in vars(inputs).items():
        if value is not None:
            kwargs[key] = value

    kwargs.pop("fct")
    kwargs.pop("x")

    if "data" in kwargs:
        kwargs["data"] = [dict(name=kwargs["data"], type="constraint", value=1.0)]

    globals()[inputs.fct](**kwargs)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
"""
A Novel Bandit-Based Approach to Hyperparameter Optimization
============================================================

Implement Hyperband to exploit configurations with fixed resource efficiently

"""
import copy
import hashlib
import logging
from collections import OrderedDict

import numpy
from tabulate import tabulate

from orion.algo.base import BaseAlgorithm
from orion.algo.space import Fidelity

logger = logging.getLogger(__name__)

REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
Hyperband cannot be used if space does not contain a fidelity dimension.
For more information on the configuration and usage of Hyperband, see
https://orion.readthedocs.io/en/develop/user/algorithms.html#hyperband
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(max_resources, reduction_factor):
    """Compute the budgets used for each execution of hyperband"""
    num_brackets = int(numpy.log(max_resources) / numpy.log(reduction_factor))
    budgets = []
    budgets_tab = {}  # just for display consideration
    for bracket_id in range(0, num_brackets + 1):
        bracket_budgets = []
        num_trials = int(
            numpy.ceil(
                int((num_brackets + 1) / (num_brackets - bracket_id + 1))
                * (reduction_factor ** (num_brackets - bracket_id))
            )
        )

        min_resources = max_resources / reduction_factor ** (num_brackets - bracket_id)
        for i in range(0, num_brackets - bracket_id + 1):
            n_i = int(num_trials / reduction_factor ** i)
            min_i = int(min_resources * reduction_factor ** i)
            bracket_budgets.append((n_i, min_i))

            if budgets_tab.get(i):
                budgets_tab[i].append((n_i, min_i))
            else:
                budgets_tab[i] = [(n_i, min_i)]

        budgets.append(bracket_budgets)

    display_budgets(budgets_tab, max_resources, reduction_factor)

    return budgets


def tabulate_status(brackets):

    header = ["i"] + ["n_i", "r_i"] * len(brackets)

    data = []

    num_rungs = max(len(bracket.rungs) for bracket in brackets)
    for rung_id in range(num_rungs):
        row = [rung_id]
        for bracket in brackets:
            if len(bracket.rungs) <= rung_id:
                row.extend(["", ""])
                continue
            in_i = len(bracket.rungs[rung_id]["results"])
            n_i = bracket.rungs[rung_id]["n_trials"]
            r_i = bracket.rungs[rung_id]["resources"]
            row.append("{:>3}/{:>3}".format(in_i, n_i))
            row.append(r_i)
        data.append(row)
    table = tabulate(data, header, tablefmt="github")
    logger.debug(table)


def display_budgets(budgets_tab, max_resources, reduction_factor):
    """Display hyperband budget as a table in debug log"""
    num_brackets = len(budgets_tab[0])
    table_str = "Display Budgets:\n"
    col_format_str = "{:<4}" + " {:<12}" * num_brackets + "\n"
    col_title_list = ["i  "] + ["n_i  r_i"] * num_brackets
    col_sub_list = ["---"] + ["---------"] * num_brackets
    table_str += col_format_str.format(*col_sub_list)
    table_str += col_format_str.format(*col_title_list)
    table_str += col_format_str.format(*col_sub_list)

    total_trials = 0
    for key, values in sorted(budgets_tab.items()):
        table_row = "{:<4} ".format(key)
        for value in values:
            n_i, r_i = value
            total_trials += n_i
            st = "{:<5} {:<7}".format(n_i, r_i)
            table_row += st
        table_str += table_row + "\n"
    table_str += col_format_str.format(*col_sub_list)
    table_str += "max resource={}, eta={}, trials number of one execution={}\n".format(
        max_resources, reduction_factor, total_trials
    )
    logger.debug(table_str)


class Hyperband(BaseAlgorithm):
    """Hyperband formulates hyperparameter optimization as a pure-exploration non-stochastic
    infinite-armed bandit problem where a predefined resource like iterations, data samples,
    or features is allocated to randomly sampled configurations.`

    For more information on the algorithm,
    see original paper at http://jmlr.org/papers/v18/16-558.html.

    Li, Lisha et al. "Hyperband: A Novel Bandit-Based Approach to Hyperparameter Optimization"
    Journal of Machine Learning Research, 18:1-52, 2018.

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    repetitions: int
        Number of execution of Hyperband. Default is numpy.inf which means to
        run Hyperband until no new trials can be suggested.

    """

    def __init__(self, space, seed=None, repetitions=numpy.inf):
        self.brackets = []
        super(Hyperband, self).__init__(space, seed=seed, repetitions=repetitions)

        self.trial_to_brackets = {}  # Stores Point id (with no fidelity) -> Bracket
        self._trials_info = {}

        fidelity_index = self.fidelity_index
        if fidelity_index is None:
            raise RuntimeError(SPACE_ERROR)

        fidelity_dim = space.values()[fidelity_index]

        self.min_resources = fidelity_dim.low
        self.max_resources = fidelity_dim.high
        self.reduction_factor = fidelity_dim.base

        # if self.reduction_factor < 2:
        #     raise AttributeError("Reduction factor for Hyperband needs to be at least 2.")

        self.repetitions = repetitions

        if self.reduction_factor >= 2:
            self.budgets = compute_budgets(self.max_resources, self.reduction_factor)
            self.brackets = self.create_brackets()
            self.seed_rng(seed)
        else:
            self.budgets = None
            self.brackets = None
            logger.warning("Reduction factor for Hyperband needs to be at least 2")

    def create_bracket(self, i, budgets, iteration):
        return HyperbandBracket(self, budgets, iteration)

    def sample_from_bracket(self, bracket, num):
        """Sample new points from bracket"""
        points = []
        while len(points) < num:
            point = bracket.get_sample()
            if point is None:
                break

            point = list(point)
            point[self.fidelity_index] = bracket.rungs[0]["resources"]

            full_id = self.get_id(point, ignore_fidelity=False)
            id_wo_fidelity = self.get_id(point, ignore_fidelity=True)
            bracket_observed = self.trial_to_brackets.get(id_wo_fidelity)

            if not self.has_suggested(point) and (
                not bracket_observed
                or (
                    bracket_observed.repetition_id < bracket.repetition_id
                    and bracket_observed.get_point_max_resource(point)
                    < bracket.rungs[0]["resources"]
                )
            ):
                # if no duplicated found or the duplicated found existing in previous hyperband
                # execution with less resource

                points.append(tuple(point))
                self.register(point)
                self.trial_to_brackets[id_wo_fidelity] = bracket

        return points

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.seed = seed
        self.rng = numpy.random.RandomState(seed)
        self.seed_brackets(seed)

    def seed_brackets(self, seed):
        rng = numpy.random.RandomState(seed)
        for i, bracket in enumerate(self.brackets[::-1]):
            bracket.seed_rng(tuple(rng.randint(0, 1000000, size=3)))

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return {
            "rng_state": self.rng.get_state(),
            "seed": self.seed,
            "_trials_info": copy.deepcopy(self._trials_info),
            "trial_to_brackets": copy.deepcopy(self.trial_to_brackets),
            "brackets": [bracket.state_dict for bracket in self.brackets],
        }

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self.seed_rng(state_dict["seed"])
        self.rng.set_state(state_dict["rng_state"])
        self._trials_info = state_dict["_trials_info"]
        self.trial_to_brackets = state_dict["trial_to_brackets"]

        while len(self.brackets) < len(state_dict["brackets"]):
            self.append_brackets()
        assert len(self.brackets) == len(state_dict["brackets"]), "corrupted state"

        for bracket, bracket_state_dict in zip(self.brackets, state_dict["brackets"]):
            bracket.set_state(bracket_state_dict)

    def register_samples(self, bracket, samples):
        for sample in samples:
            full_id = self.get_id(sample, ignore_fidelity=False)
            if self.has_observed(sample):
                raise RuntimeError(
                    "Hyperband resampling a trial that was already completed. "
                    "This should never happen. "
                    "If you get this error please report this issue on github at "
                    "https://github.com/Epistimio/orion/issues/new/choose"
                )
            self.register(sample)
            bracket.register(sample, None)

            if self.get_id(sample, ignore_fidelity=True) not in self.trial_to_brackets:
                self.trial_to_brackets[
                    self.get_id(sample, ignore_fidelity=True)
                ] = bracket

    def promote(self, num):
        samples = []
        for bracket in reversed(self.brackets):
            if bracket.is_ready() and not bracket.is_done:
                bracket_samples = bracket.promote(num - len(samples))
                self.register_samples(bracket, bracket_samples)
                samples += bracket_samples

        return samples

    def sample(self, num):
        samples = []
        for bracket in reversed(self.brackets):
            if not bracket.is_filled:
                bracket_samples = self.sample_from_bracket(
                    bracket, min(num - len(samples), bracket.remainings)
                )
                self.register_samples(bracket, bracket_samples)
                samples += bracket_samples

        return samples

    def suggest(self, num):
        """Suggest a number of new sets of parameters.

        Sample new points until first rung is filled. Afterwards
        waits for all trials to be completed before promoting trials
        to the next rung.

        Parameters
        ----------
        num: int, optional
            Number of points to suggest. Defaults to 1.

        Returns
        -------
        list of points or None
            A list of lists representing points suggested by the algorithm. The algorithm may opt
            out if it cannot make a good suggestion at the moment (it may be waiting for other
            trials to complete), in which case it will return None.

        """
        self._refresh_brackets()

        samples = self.promote(num)

        samples.extend(self.sample(max(num - len(samples), 0)))

        tabulate_status(self.brackets)

        if samples:
            return samples

        # Either all brackets are done or none are ready and algo needs to wait for some trials to
        # complete
        if len(self.trial_to_brackets) >= self.space.cardinality:
            logger.warning(
                "The number of unique trials of bottom rungs exceeds the search space "
                "cardinality %i, Hyperband algorithm exits.",
                self.space.cardinality,
            )
        else:
            logger.warning(
                f"{self.__class__.__name__} cannot suggest new samples, exit."
            )

        return []

    @property
    def executed_times(self):
        """Counter for how many times Hyperband been executed"""
        if not self.brackets:
            return 0
        executed_times = self.brackets[0].repetition_id
        return executed_times - int(not self.brackets[0].is_done)

    def _refresh_brackets(self):
        """Refresh bracket if one hyperband execution is done"""
        if all(bracket.is_done for bracket in self.brackets):
            logger.debug(
                "Hyperband execution %i is done, required to execute %s times",
                self.executed_times,
                str(self.repetitions),
            )

            # Continue to the next execution if need
            if self.executed_times < self.repetitions:
                self.append_brackets()

    def append_brackets(self):
        self.brackets = self.create_brackets() + self.brackets
        # Reset brackets seeds
        self.seed_brackets(self.seed)

    def create_brackets(self):
        return [
            self.create_bracket(i, bracket_budgets, self.executed_times + 1)
            for i, bracket_budgets in enumerate(self.budgets)
        ]

    def _get_bracket(self, point):
        """Get the bracket of a point"""
        fidelity = point[self.fidelity_index]
        _id_wo_fidelity = self.get_id(point, ignore_fidelity=True)

        brackets = []
        for bracket in self.brackets:
            # If find same point in first rung of a bracket,
            # the point should register in this bracket
            if _id_wo_fidelity in bracket.rungs[0]["results"]:
                brackets = [bracket]
                break

        if not brackets:
            # If the point show in current hyeprband execution the first time,
            # the bracket with same fidelity in the first rung should be used,
            # the assumption is that there is no duplicated points inside same hyperband execution.
            brackets = [
                bracket
                for bracket in self.brackets
                if bracket.rungs[0]["resources"] == fidelity
            ]

        if not brackets:
            raise ValueError(
                "No bracket found for point {0} with fidelity {1}".format(
                    _id_wo_fidelity, fidelity
                )
            )

        if len(brackets) > 1:
            logger.warning(
                "More than one bracket found for point %s, this should not happen",
                str(point),
            )

        bracket = brackets[0]

        return bracket

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        A simple random sampler though does not take anything into account.
        """

        for point, result in zip(points, results):

            if not self.has_suggested(point):
                logger.info(
                    "Ignoring point %s because it was not sampled by current algo.",
                    point,
                )
                continue

            self.register(point, result)

            bracket = self._get_bracket(point)

            try:
                bracket.register(point, result["objective"])
            except IndexError:
                logger.warning(
                    "Point registered to wrong bracket. This is likely due "
                    "to a corrupted database, where trials of different fidelity "
                    "have a wrong timestamps."
                )
                continue

    @property
    def is_done(self):
        """Return True, if all required execution been done."""
        if self.executed_times >= self.repetitions:
            return True
        return False


class HyperbandBracket:
    """Bracket of rungs for the algorithm Hyperband.

    Parameters
    ----------
    hyperband: `Hyperband` algorithm
        The hyperband algorithm object which this bracket will be part of.
    budgets: list of tuple
        Each tuple gives the (n_trials, resource_budget) for the respective rung.
    repetition_id: int
        The id of hyperband execution this bracket belongs to

    """

    def __init__(self, hyperband, budgets, repetition_id):
        self.hyperband = hyperband
        self.rungs = [
            dict(resources=budget, n_trials=n_trials, results=OrderedDict())
            for n_trials, budget in budgets
        ]
        self.seed = None
        self.repetition_id = repetition_id
        self.buffer = 10
        self._samples = None

        logger.debug("Bracket budgets: %s", str(budgets))

        # points = hyperband.sample(compute_rung_sizes(reduction_factor, len(budgets))[0])
        # for point in points:
        #     self.register(point, None)

    @property
    def state_dict(self):
        return {"rungs": copy.deepcopy(self.rungs)}

    def set_state(self, state_dict):
        self.rungs = state_dict["rungs"]

    @property
    def is_filled(self):
        """Return True if first rung with trials is filled"""
        return self.has_rung_filled(0)

    def get_point_max_resource(self, point):
        """Return the max resource value that has been tried for a point"""
        max_resource = 0
        _id_wo_fidelity = self.hyperband.get_id(point, ignore_fidelity=True)
        for rung in self.rungs:
            if _id_wo_fidelity in rung["results"]:
                max_resource = rung["resources"]

        return max_resource

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, seed):
        self._seed = seed
        self._samples = None

    def get_sample(self):
        if self._samples is None:
            n_samples = int(self.rungs[0]["n_trials"] * self.buffer)
            self._samples = self.hyperband.space.sample(n_samples, seed=self.seed)

        return self._samples.pop(0) if self._samples else None

    def sample(self, num):
        """Sample a new trial with lowest fidelity"""
        should_have_n_trials = self.rungs[0]["n_trials"]
        n_trials = len(self.rungs[0]["results"])
        request = max(min(should_have_n_trials - n_trials, num), 0)
        if request == 0:
            return []

        return self.hyperband.sample_for_bracket(
            request, self, buffer=should_have_n_trials * 10 / request
        )

    def register(self, point, objective):
        """Register a point in the corresponding rung"""
        self._get_results(point)[self.hyperband.get_id(point, ignore_fidelity=True)] = (
            objective,
            point,
        )

    def _get_results(self, point):
        fidelity = point[self.hyperband.fidelity_index]
        rungs = [
            rung["results"] for rung in self.rungs if rung["resources"] == fidelity
        ]
        if not rungs:
            budgets = [rung["resources"] for rung in self.rungs]
            raise IndexError(
                REGISTRATION_ERROR.format(
                    fidelity=fidelity, budgets=budgets, params=point
                )
            )

        return rungs[0]

    @property
    def remainings(self):
        should_have_n_trials = self.rungs[0]["n_trials"]
        have_n_trials = len(self.rungs[0]["results"])
        return max(should_have_n_trials - have_n_trials, 0)

    def get_candidates(self, rung_id):
        """Get a candidate for promotion"""
        if self.has_rung_filled(rung_id + 1):
            return []

        rung = self.rungs[rung_id]["results"]
        next_rung = self.rungs[rung_id + 1]["results"]

        rung = list(sorted((objective, point) for objective, point in rung.values()))

        should_have_n_trials = self.rungs[rung_id + 1]["n_trials"]
        points = []
        i = 0
        while len(points) + len(next_rung) < should_have_n_trials:
            objective, point = rung[i]
            assert objective is not None
            _id = self.hyperband.get_id(point, ignore_fidelity=True)
            if _id not in next_rung:
                points.append(point)
            i += 1

        return points

    @property
    def is_done(self):
        """Return True, if the last rung is filled."""
        return self.has_rung_filled(len(self.rungs) - 1)

    def has_rung_filled(self, rung_id):
        """Return True, if the rung[rung_id] is filled."""
        n_trials = len(self.rungs[rung_id]["results"])
        return n_trials >= self.rungs[rung_id]["n_trials"]

    def is_ready(self, rung_id=None):
        """Return True, if the bracket is ready for next promote"""
        if rung_id is not None:
            return self.has_rung_filled(rung_id) and all(
                objective is not None
                for objective, _ in self.rungs[rung_id]["results"].values()
            )

        is_ready = False
        for _rung_id in range(len(self.rungs)):
            if self.has_rung_filled(_rung_id):
                is_ready = self.is_ready(_rung_id)
            else:
                break

        return is_ready

    def promote(self, num):
        """Promote the first candidate that is found and return it

        The rungs are iterated over in reversed order, so that high rungs
        are prioritised for promotions. When a candidate is promoted, the loop is broken and
        the method returns the promoted point.

        .. note ::

            All trials are part of the rungs, for any state. Only completed trials
            are eligible for promotion, i.e., only completed trials can be part of top-k.
            Lookup for promotion in rung l + 1 contains trials of any status.

        """
        if self.is_done:
            return []

        for rung_id in range(len(self.rungs)):
            # No more promotion possible, skip to next rung
            if self.has_rung_filled(rung_id + 1):
                continue

            if not self.is_ready(rung_id):
                return []

            points = []
            for candidate in self.get_candidates(rung_id):
                # pylint: disable=logging-format-interpolation
                logger.debug(
                    "Promoting {point} from rung {past_rung} with fidelity {past_fidelity} to "
                    "rung {new_rung} with fidelity {new_fidelity}".format(
                        point=candidate,
                        past_rung=rung_id,
                        past_fidelity=candidate[self.hyperband.fidelity_index],
                        new_rung=rung_id + 1,
                        new_fidelity=self.rungs[rung_id + 1]["resources"],
                    )
                )

                candidate = list(copy.deepcopy(candidate))
                candidate[self.hyperband.fidelity_index] = self.rungs[rung_id + 1][
                    "resources"
                ]
                if not self.hyperband.has_suggested(candidate):
                    points.append(tuple(candidate))

            return points[:num]

        return []

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.seed = seed

    def __repr__(self):
        """Return representation of bracket with fidelity levels"""
        return "{}(resource={}, repetition id={})".format(
            self.__class__.__name__,
            [rung["resources"] for rung in self.rungs],
            self.repetition_id,
        )



# -*- coding: utf-8 -*-
"""
Asynchronous Successive Halving Algorithm
=========================================
"""
import copy
import hashlib
import logging
from collections import defaultdict

import numpy

from orion.algo.base import BaseAlgorithm
from orion.algo.hyperband import Hyperband, HyperbandBracket, display_budgets
from orion.algo.space import Fidelity

logger = logging.getLogger(__name__)


REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
ASHA can only be used if there is one fidelity dimension.
For more information on the configuration and usage of ASHA, see
https://orion.readthedocs.io/en/develop/user/algorithms.html#asha
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(
    min_resources, max_resources, reduction_factor, num_rungs, num_brackets
):
    """Compute the budgets used for ASHA"""

    # Tracks state for new trial add
    if num_brackets > num_rungs:
        logger.warning(
            "The input num_brackets %i is larger than the number of rungs %i, "
            "set num_brackets as %i",
            num_brackets,
            num_rungs,
            num_rungs,
        )
        num_brackets = num_rungs

    budgets = numpy.logspace(
        numpy.log(min_resources) / numpy.log(reduction_factor),
        numpy.log(max_resources) / numpy.log(reduction_factor),
        num_rungs,
        base=reduction_factor,
    )
    budgets = (budgets + 0.5).astype(int)

    for i in range(num_rungs - 1):
        if budgets[i] >= budgets[i + 1]:
            budgets[i + 1] = budgets[i] + 1

    if budgets[-1] > max_resources:
        raise ValueError(BUDGET_ERROR.format(min_resources, max_resources, num_rungs))

    budgets = [budgets[bracket_index:] for bracket_index in range(num_brackets)]

    ressources = budgets
    budgets = []
    budgets_tab = defaultdict(list)
    for bracket_ressources in ressources:
        bracket_budgets = []
        for i, min_ressources in enumerate(bracket_ressources[::-1]):
            budget = (reduction_factor ** i, min_ressources)
            bracket_budgets.append(budget)
            budgets_tab[len(bracket_ressources) - i - 1].append(budget)
        budgets.append(bracket_budgets[::-1])

    display_budgets(budgets_tab, max_resources, reduction_factor)

    return list(budgets)


class ASHA(Hyperband):
    """Asynchronous Successive Halving Algorithm

    `A simple and robust hyperparameter tuning algorithm with solid theoretical underpinnings
    that exploits parallelism and aggressive early-stopping.`

    For more information on the algorithm, see original paper at https://arxiv.org/abs/1810.05934.

    Li, Liam, et al. "Massively parallel hyperparameter tuning."
    arXiv preprint arXiv:1810.05934 (2018)

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    num_rungs: int, optional
        Number of rungs for the largest bracket. If not defined, it will be equal to (base + 1) of
        the fidelity dimension. In the original paper,
        num_rungs == log(fidelity.high/fidelity.low) / log(fidelity.base) + 1.
        Default: log(fidelity.high/fidelity.low) / log(fidelity.base) + 1
    num_brackets: int
        Using a grace period that is too small may bias ASHA too strongly towards
        fast converging trials that do not lead to best results at convergence (stagglers). To
        overcome this, you can increase the number of brackets, which increases the amount of
        resource required for optimisation but decreases the bias towards stragglers.
        Default: 1
    repetitions: int
        Number of execution of ASHA. Default is numpy.inf which means to
        run ASHA until no new trials can be suggested.

    """

    def __init__(
        self,
        space,
        seed=None,
        num_rungs=None,
        num_brackets=1,
        repetitions=numpy.inf,
    ):
        super(ASHA, self).__init__(space, seed=seed, repetitions=repetitions)

        self.num_rungs = num_rungs
        self.num_brackets = num_brackets

        self._param_names += ["num_rungs", "num_brackets"]

        if self.reduction_factor < 2:
            raise AttributeError("Reduction factor for ASHA needs to be at least 2.")

        if num_rungs is None:
            num_rungs = int(
                numpy.log(self.max_resources / self.min_resources)
                / numpy.log(self.reduction_factor)
                + 1
            )

        self.num_rungs = num_rungs

        self.budgets = compute_budgets(
            self.min_resources,
            self.max_resources,
            self.reduction_factor,
            self.num_rungs,
            self.num_brackets,
        )

        self.brackets = self.create_brackets()

    def compute_bracket_idx(self, num):
        def assign_resources(n, remainings, totals):
            if n == 0 or remainings.sum() == 0:
                return remainings

            ratios = remainings / totals
            fractions = numpy.nan_to_num(ratios / ratios.sum(), 0)
            index = numpy.argmax(fractions)
            remainings[index] -= 1
            n -= 1
            if n > 0:
                return assign_resources(n, remainings, totals)

            return remainings

        remainings = numpy.array([bracket.remainings for bracket in self.brackets])
        totals = numpy.array(
            [bracket.rungs[0]["n_trials"] for bracket in self.brackets]
        )
        remainings_after = copy.deepcopy(remainings)
        assign_resources(num, remainings_after, totals)
        allocations = remainings - remainings_after

        return allocations

    def sample(self, num):
        samples = []
        bracket_nums = self.compute_bracket_idx(num)
        for idx, bracket_num in enumerate(bracket_nums):
            bracket = self.brackets[idx]
            bracket_samples = self.sample_from_bracket(bracket, bracket_num)
            self.register_samples(bracket, bracket_samples)
            samples += bracket_samples

        return samples

    def suggest(self, num):
        return super(ASHA, self).suggest(1)

    def create_bracket(self, i, budgets, iteration):
        return ASHABracket(self, budgets, iteration)


class ASHABracket(HyperbandBracket):
    """Bracket of rungs for the algorithm ASHA.

    Parameters
    ----------
    asha: `ASHA` algorithm
        The asha algorithm object which this bracket will be part of.
    budgets: list of tuple
        Each tuple gives the (n_trials, resource_budget) for the respective rung.
    repetition_id: int
        The id of hyperband execution this bracket belongs to

    """

    def sample(self, num):
        """Sample a new trial with lowest fidelity"""
        should_have_n_trials = self.rungs[0]["n_trials"]
        return self.hyperband.sample_for_bracket(num, self)

    def get_candidate(self, rung_id):
        """Get a candidate for promotion"""
        rung = self.rungs[rung_id]["results"]
        next_rung = self.rungs[rung_id + 1]["results"]

        rung = list(
            sorted(
                (objective, point)
                for objective, point in rung.values()
                if objective is not None
            )
        )
        k = len(rung) // self.hyperband.reduction_factor
        k = min(k, len(rung))

        for i in range(k):
            point = rung[i][1]
            _id = self.hyperband.get_id(point, ignore_fidelity=True)
            if _id not in next_rung:
                return point

        return None

    @property
    def is_filled(self):
        """ASHA's first rung can always sample new points"""
        return False

    def is_ready(self, rung_id=None):
        """ASHA's always ready for promotions"""
        return True

    def promote(self, num):
        """Promote the first candidate that is found and return it

        The rungs are iterated over in reversed order, so that high rungs
        are prioritised for promotions. When a candidate is promoted, the loop is broken and
        the method returns the promoted point.

        .. note ::

            All trials are part of the rungs, for any state. Only completed trials
            are eligible for promotion, i.e., only completed trials can be part of top-k.
            Lookup for promotion in rung l + 1 contains trials of any status.

        """
        if num < 1 or self.is_done:
            return []

        for rung_id in range(len(self.rungs) - 2, -1, -1):
            candidate = self.get_candidate(rung_id)
            if candidate:

                # pylint: disable=logging-format-interpolation
                logger.debug(
                    "Promoting {point} from rung {past_rung} with fidelity {past_fidelity} to "
                    "rung {new_rung} with fidelity {new_fidelity}".format(
                        point=candidate,
                        past_rung=rung_id,
                        past_fidelity=candidate[self.hyperband.fidelity_index],
                        new_rung=rung_id + 1,
                        new_fidelity=self.rungs[rung_id + 1]["resources"],
                    )
                )

                candidate = list(copy.deepcopy(candidate))
                candidate[self.hyperband.fidelity_index] = self.rungs[rung_id + 1][
                    "resources"
                ]

                return [tuple(candidate)]

        return []



# -*- coding: utf-8 -*-
"""
Mocks Oríon's runtime
=====================

Boilerplate to simulate Oríon's runtime and data sources.

"""
# pylint: disable=protected-access

import os
import tempfile

import yaml

from orion.core.io import experiment_builder as experiment_builder
from orion.core.utils.singleton import (
    SingletonAlreadyInstantiatedError,
    update_singletons,
)
from orion.core.worker.trial import Trial
from orion.storage.base import Storage, get_storage


# pylint: disable=no-self-use,protected-access
class BaseOrionState:
    """Setup global variables and singleton for tests.

    It swaps the singleton with `None` at startup and restores them after the tests.
    It also initializes PickleDB as the storage for testing.
    We use PickledDB as our storage mock

    Parameters
    ----------
    benchmarks: List, optional
       List of benchmarks ot insert into the database
    experiments: list, optional
        List of experiments to insert into the database
    trials: list, optional
        List of trials to insert into the database
    workers: list, optional
        List of workers to insert into the database
    lies: list, optional
        List of lies to insert into the database
    resources: list, optional
        List of resources to insert into the database
    from_yaml: YAML, optional
        YAML config to apply for this test
    storage: dict, optional
        Configuration of the underlying storage backend

    Examples
    --------
    >>> my_config = {...}
    >>> with OrionState(my_config):
        ...

    """

    # TODO: Fix these singletons to remove Legacy, MongoDB, PickledDB and EphemeralDB.
    singletons = {}
    experiments = []
    trials = []
    resources = []
    workers = []

    def __init__(
        self,
        benchmarks=None,
        experiments=None,
        trials=None,
        workers=None,
        lies=None,
        resources=None,
        from_yaml=None,
        storage=None,
    ):
        if from_yaml is not None:
            with open(from_yaml) as f:
                exp_config = list(yaml.safe_load_all(f))
                experiments = exp_config[0]
                trials = exp_config[1]

        self.tempfile = None
        self.tempfile_path = None
        self.storage_config = _select(storage, _get_default_test_storage())

        self._benchmarks = _select(benchmarks, [])
        self._experiments = _select(experiments, [])
        self._trials = _select(trials, [])
        self._workers = _select(workers, [])
        self._resources = _select(resources, [])
        self._lies = _select(lies, [])

        # In case of track we also store the inserted object
        # so the user can compare in tests the different values
        self.trials = []
        self.experiments = self._experiments
        self.lies = []

    def init(self, config):
        """Initialize environment before testing"""
        self.storage(config)
        self.load_experience_configuration()
        return self

    def get_experiment(self, name, version=None):
        """Make experiment id deterministic"""
        exp = experiment_builder.build(name=name, version=version)
        return exp

    def get_trial(self, index):
        """Return a Trial"""
        return Trial(**self.trials[index])

    def cleanup(self):
        """Cleanup after testing"""
        if self.tempfile is not None:
            os.close(self.tempfile)
        _remove(self.tempfile_path)

    def _set_tables(self):
        self.trials = []
        self.lies = []

        for exp in self._experiments:
            get_storage().create_experiment(exp)

        for trial in self._trials:
            nt = get_storage().register_trial(Trial(**trial))
            self.trials.append(nt.to_dict())

        for lie in self._lies:
            nt = get_storage().register_lie(Trial(**lie))
            self.lies.append(nt.to_dict())

    def load_experience_configuration(self):
        """Load an example database."""
        for i, t_dict in enumerate(self._trials):
            self._trials[i] = Trial(**t_dict).to_dict()

        for i, t_dict in enumerate(self._lies):
            self._lies[i] = Trial(**t_dict).to_dict()

        self._trials.sort(key=lambda obj: int(obj["_id"], 16), reverse=True)

        for i, experiment in enumerate(self._experiments):
            if "user_script" in experiment["metadata"]:
                path = os.path.join(
                    os.path.dirname(__file__), experiment["metadata"]["user_script"]
                )
                experiment["metadata"]["user_script"] = path

            experiment["_id"] = i

        self._set_tables()

    def make_config(self):
        """Iterate over the database configuration and replace ${file}
        by the name of a temporary file
        """
        self.tempfile, self.tempfile_path = tempfile.mkstemp("_orion_test")
        _remove(self.tempfile_path)

        def map_dict(fun, dictionary):
            """Return a dictionary with fun applied to each values"""
            return {k: fun(v) for k, v in dictionary.items()}

        def replace_file(v):
            """Replace `${file}` by a generated temporary file"""
            if isinstance(v, str):
                v = v.replace("${file}", self.tempfile_path)

            if isinstance(v, dict):
                v = map_dict(replace_file, v)

            return v

        return map_dict(replace_file, self.storage_config)

    def __enter__(self):
        """Load a new database state"""
        self.singletons = update_singletons()
        self.cleanup()
        return self.init(self.make_config())

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup database state"""
        self.cleanup()

        update_singletons(self.singletons)

    def storage(self, config=None):
        """Return test storage"""
        if config is None:
            return get_storage()

        try:
            config["of_type"] = config.pop("type")
            db = Storage(**config)
            self.storage_config = config
        except SingletonAlreadyInstantiatedError:
            db = get_storage()

        except KeyError:
            print(self.storage_config)
            raise

        return db


class LegacyOrionState(BaseOrionState):
    """See :func:`~orion.testing.state.BaseOrionState`"""

    def __init__(self, *args, **kwargs):
        super(LegacyOrionState, self).__init__(*args, **kwargs)
        self.initialized = False

    @property
    def database(self):
        """Retrieve legacy database handle"""
        return get_storage()._db

    def init(self, config):
        """Initialize environment before testing"""
        self.storage(config)
        self.initialized = True

        if hasattr(get_storage(), "_db"):
            self.database.remove("experiments", {})
            self.database.remove("trials", {})

        self.load_experience_configuration()
        return self

    def get_experiment(self, name, version=None):
        """Make experiment id deterministic"""
        exp = experiment_builder.build(name, version=version)
        exp._id = exp.name
        return exp

    def _set_tables(self):
        if self._benchmarks:
            self.database.write("benchmarks", self._benchmarks)
        if self._experiments:
            self.database.write("experiments", self._experiments)
        if self._trials:
            self.database.write("trials", self._trials)
        if self._workers:
            self.database.write("workers", self._workers)
        if self._resources:
            self.database.write("resources", self._resources)
        if self._lies:
            self.database.write("lying_trials", self._lies)

        self.lies = self._lies
        self.trials = self._trials

    def cleanup(self):
        """Cleanup after testing"""
        if self.initialized:
            self.database.remove("benchmarks", {})
            self.database.remove("experiments", {})
            self.database.remove("trials", {})
            if self.tempfile is not None:
                os.close(self.tempfile)
            _remove(self.tempfile_path)
        self.initialized = False


# We are faking a class constructor here
# pylint: disable=C0103
def OrionState(*args, **kwargs):
    """Build an orion state in function of the storage type"""
    storage = kwargs.get("storage")

    if not storage or storage["type"] == "legacy":
        return LegacyOrionState(*args, **kwargs)

    return BaseOrionState(*args, **kwargs)


def _get_default_test_storage():
    """Return default configuration for the test storage"""
    return {"type": "legacy", "database": {"type": "PickledDB", "host": "${file}"}}


def _remove(file_name):
    if file_name is None:
        return

    try:
        os.remove(file_name)
    except FileNotFoundError:
        pass


def _select(lhs, rhs):
    if lhs:
        return lhs
    return rhs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.algo.random`."""
import numpy
import pytest

from orion.algo.random import Random
from orion.algo.space import Integer, Real, Space
from orion.testing.algo import BaseAlgoTests


class TestRandomSearch(BaseAlgoTests):
    algo_name = "random"
    config = {"seed": 123456}


TestRandomSearch.set_phases([("random", 0, "space.sample")])
